"""
Training a network happens here
"""

import numpy as np
from datetime import datetime
import os
import copy

import logging
logger = logging.getLogger(__name__)

from . import layer
from . import utils
from . import err
from . import act
from . import plot
from . import regul
from . import opt



class Training:
	"""
	Holds together everthing related to the process of training a Net (or a WNet): the training data and the network.
	"""

	
	def __init__(self, net, dat, errfctname="msb", regulweight=None, regulfctname=None, itersavepath=None, saveeachit=False, autoplotdirpath=".", autoplot=False, logpath=None, trackbiases=False, verbose=False, name=None):
		"""

		Sets up
		- housekeeping lists
		- error function
		
		:param dat: can be None, if you want a dummy container for your Net.
		
		:param trackbiases: If True, will track and plot the evolution of biases as function of input feature values.
			Could get massive and slow down the training. Also the plot is rather slow.
		:type trackbiases: bool
		
		:param logpath: If not None, I will log to this file, and not propagate my logger at all.
			Unfortunately this does not work well when using multiple threads/processes,
			as logs from every thread get written into the logfiles.
		
		:param saveeachit: If True, writes the full status and history to disk at each iteration.
			If False, only some snapshots are written, but they still contain the full history!
		:type saveeachtit: bool
		
		
		"""
		
		# We start by dealing with the logger.
		if logpath is not None:
			fh = logging.FileHandler(logpath, delay=True)
			fh.setLevel(logging.DEBUG)
			fh.setFormatter(logging.Formatter("PID %(process)d: %(levelname)s: %(name)s(%(funcName)s): %(message)s"))
			logger.addHandler(fh)
			logger.propagate = False
			# Even if the file is only openend at the first message (delay=True in FileHandler above),
			# we do have to make sure the directories exist, so that logging can be done at any time.
			logpathdir = os.path.split(logpath)[0]
			if not os.path.exists(logpathdir):
				os.makedirs(logpathdir)
			
		
		self.name = name
	
		self.set_net(net)
		self.set_dat(dat)
		
		# Let's check compatibility between those two!
		if dat != None:
			assert net.ni == self.dat.getni()
			if errfctname in ["mse", "msb", "msrb", "msre", "msbw", "msbwsignorm"]:
				assert net.no == self.dat.getno()
			elif errfctname in ["msbwnet"]:
				assert net.no == 2*self.dat.getno()
			else:
				logger.warning("Unknown error function, will blindly go ahead...")
				
		# Setting up the cost function
		self.errfctname = errfctname
		self.errfct = eval("err.{0}".format(self.errfctname))
		if regulfctname is not None:
			if regulweight is None:
				raise ValueError("Regularisation is set, but no weight")
			else:
				if regulweight == 0:
					logger.warning("Regularisation is set, but the weight is zero! Let's boldy go ahead anyhow.")
				elif regulweight < 0:
					logger.warning("Regularisation is set, but the weight is negative! Let's boldy go ahead anyhow.")
				self.regulfctname = regulfctname
				self.regulfct = eval("regul.{0}".format(self.regulfctname))
				self.regullam = regulweight
		else:
			self.regullam = None
			
		# We initialize some counters for the optimization:
		self.optit = 0 # The iteration counter
		self.optcall = 0 # The cost function call counter
		self.optitcall = 0 # Idem, but gets reset at each new iteration
		self.opterr = np.inf # The current cost function value on the training set
		
		# And some lists describing the optimization:
		self.opterrs = [] # The cost function value on the training set at each (!) call
		
		self.optitparams = [] # A copy of the network parameters at each iteration
		self.optiterrs_train = [] # The cost function value on the training set at each iteration
		self.optiterrs_val = [] # The cost function value on the validation set at each iteration

		self.optitcalls = [] # The cost function call counter at each iteration
		self.optittimes = [] # Time taken for iteration, in seconds
		self.optbatchchangeits = [] # Iteration counter (in fact indices) when the batche gets changed
		
		self.trackbiases = trackbiases # Should we save all the biases at each snapshot (e.g. minibatch-change) (and make the related plots) ?
		self.biassnaps_train = []
		self.biassnaps_val = []
		self.biassnaps_it = [] # Iteration counter of snapshot
		
		self.verbose = verbose
		self.itersavepath = itersavepath
		self.saveeachit = saveeachit
		
		self.autoplotdirpath = autoplotdirpath
		self.autoplot = autoplot
		
		if dat is not None: # Otherwise it's just to create a dummy container, no need to report
			logger.info("Done with setup of {self}".format(self=self))
		
		# And let's test this out before we start, so that it fails fast in case of a problem.
		# But we only do so if the file does not yet exist, to avoid overwriting an existing training.
		# While this idea was OK for debugging, it's just weird to save anything to disk at this stage, so let's remove it.
		# This should be checked when starting the training.
		#if self.itersavepath is not None:
		#	if not os.path.exists(self.itersavepath):
		#		self.save(self.itersavepath)
	
	
	def set_dat(self, dat):
		"""
		Allows to add or replace training data (e.g. when reading a self.save()...)
		"""
		self.dat = dat
		self.set_datstr()
		

	def set_datstr(self):
		"""
		We do this so that we can still display this string on plots if dat has been removed.
		"""
		self.datstr = str(self.dat)
	

	def set_net(self, net):
		"""
		Replaces the network object
		"""
		self.net = net
		self.params = self.net.get_params_ref() # Fast connection to the network parameters
		self.paramslice = slice(None) # By default, all params are free to be optimized
		
	def set_paramslice(self, **kwargs):
		"""
		The paramslice allows to specify which params you want to be optimized.
		This is relevant for instance when training a MultNet or a WNet.
		We use a slice of this. Indexing with a boolean array ("mask") would seem nicer, but fancy indexing does not preserve
		the references. Hence using a slice is a good compromise for speed.
		
		The Net object knows how to get such a slice, given the kwargs and the pecularities of the Net.
		"""
		
		self.paramslice = self.net.get_paramslice(**kwargs)
		logger.info("Paramslice is set with kwargs '{}' : {}/{} params are free to be optimized.".format(
			kwargs, len(self.params[self.paramslice]), self.net.nparams())
			)
		

	def __str__(self):
		"""
		A short spaceless automatic description
		"""	
		if self.dat is not None: # If available, we use the live dat.
			datstr = str(self.dat)
		else: # Otherwise, we use datstr as backup solution.
			datstr = getattr(self, "datstr", "Ouch!") # To ensure that it works also if datstr is missing (backwards compatibility).

		autotxt = "{ef}({self.net}, {datstr})".format(ef=self.get_costfctname(), self=self, datstr=datstr)				
		return autotxt
	
	def get_costfctname(self):
		"""
		Returns a string with a description of the cost function
		"""
		txterrfct = "{self.errfctname}".format(self=self)
		if hasattr(self, 'regullam') and self.regullam is not None:
			txterrfct += "+{self.regullam:1.0e}x{self.regulfctname}".format(self=self)
		return txterrfct
	
	
	def title(self):
		"""
		Returns the name and string, typically nicer for plots.
		"""
		
		
		if self.name is not None:
			return "Training '{name}': {auto} ({it} it, {tmin:.1f} min)".format(name=self.name, auto=str(self), it=self.optit, tmin=np.sum(self.optittimes)/60.0)
		else:
			return "{auto} ({it} it, {tmin:.1f} min)".format(auto=str(self), it=self.optit, tmin=np.sum(self.optittimes)/60.0)
	
	
	def takeover(self, othertrain):
		"""
		This copies the net and all the progress counters and logs from another train object into the current one.
		Useful if you want to carry on a training with a new train object, typically with different settings and
		maybe on different data.
		"""
		
		logger.info("Setting up training '{}' to take over the work from '{}'...".format(self.name, othertrain.name))
		
		if not self.net.nparams() == othertrain.net.nparams():
			raise RuntimeError("Other network is not compatible, this is fishy!")
		
		self.set_net(othertrain.net)
		
		# Copying the counter positions:
		self.optit = othertrain.optit
		self.optcall = othertrain.optcall
		self.optitcall = 0 # Gets reset at each new iteration
		self.opterr = othertrain.opterr
		
		# And some lists describing the optimization:
		self.opterrs = othertrain.opterrs[:]
		
		self.optitparams = othertrain.optitparams[:]
		self.optiterrs_train = othertrain.optiterrs_train[:]
		self.optiterrs_val = othertrain.optiterrs_val[:]

		self.optitcalls = othertrain.optitcalls[:]
		self.optittimes = othertrain.optittimes[:]
		try:
			self.optbatchchangeits = othertrain.optbatchchangeits[:]
		except AttributeError:
			self.optbatchchangeits = []
		
		self.biassnaps_train = othertrain.biassnaps_train
		self.biassnaps_val = othertrain.biassnaps_val
		self.biassnaps_it = othertrain.biassnaps_it

		logger.info("Done with the takeover")
		
	

	def save(self, filepath, keepdata=False, resetcache=True):
		"""
		Saves the training progress into a pkl file
		As the training data is so massive, by default we do not save it!
		Note that this might be done at each iteration!
		"""
		if resetcache is True:
			self.net.resetcache()
		if keepdata is True:
			logger.info("Writing training to disk and keeping the data...")
			utils.writepickle(self, filepath)	
		else:
			tmptraindata = self.dat
			self.set_datstr()
			self.dat = None
			utils.writepickle(self, filepath)		
			self.dat = tmptraindata



	def plotpath(self, plotname, suffix="_optitXXXXX", dirpath=None):
		"""
		Prepares a path to be used by checkplots.
		"""
		
		if dirpath is None:
			dirpath = self.autoplotdirpath
		
		if suffix == "_optitXXXXX":
			suffix = "_optit{0:05d}".format(self.optit)
		
		return os.path.join(dirpath, plotname+suffix+".png")
		

	def makeplots(self, **kwargs):
		"""
		Saves a bunch of default checkplots into the specified directory.
		Can typically be called at the end of training, or after iterations.
		
		kwargs are passed to plotpath.
		"""
		
		logger.info("Making and writing plots...")
		plot.sumevo(self, filepath=self.plotpath("sumevo", **kwargs))
		plot.outdistribs(self, filepath=self.plotpath("outdistribs", **kwargs))
		plot.errorinputs(self, filepath=self.plotpath("errorinputs", **kwargs))
		self.plot_netviz(**kwargs)	
		if self.trackbiases:
			plot.biasevo(self, filepath=self.plotpath("biasevo", **kwargs))
		
		logger.info("Done with plots")

	def plot_netviz(self, filepath='None', **kwargs):
		"""
		Slightly different from the other plots as it takes a Net, not a Training.
		"""
		
		#write_mode = False
		#for l in self.net.layers:
		#	if not l.mode == "sum":
		#		write_mode = True
		#
		#if write_mode:
		#	title = r"$\to$".join([r"${m}^{{{n}}}\mathrm{{{actfct}}}$".format(n=l.nn, m=plot.get_symbol(l.mode, False), actfct=l.actfct.__name__) for l in self.net.layers])
		#else:
		#	title = r"$\to$".join([r"$\mathrm{{{n}/{actfct}}}$".format(n=l.nn, actfct=l.actfct.__name__) for l in self.net.layers])
		#title += r"$;\ \mathrm{{{}}}$".format(self.get_costfctname())
		#title += r"$;\ \mathrm{{it:{:05d}}}$".format(self.optit)
		#title += r"$;\ \mathrm{{{tmin:.1f}min}}$".format(tmin=np.sum(self.optittimes)/60.0)
		
		title = str(self) + ", {} it, {:.1f} min".format(self.optit, np.sum(self.optittimes)/60.0)
		
		if filepath is 'None':
			filepath = self.plotpath("netviz")
		plot.netviz(self.net, filepath=filepath, title=title, **kwargs)	
		
	def start(self):
		"""
		Called a the beginning of a training 
		"""
		logger.info("Starting an optimization cycle...")
		self.testcost()
		self.iterationstarttime = datetime.now()
		self.optitcall = 0
		
		# If this is the very first start, we prepare plots to viz the initial contidions:
		if self.optit == 0 and self.autoplot:
			self.plot_netviz()
			
			
		
	
	def end(self):
		"""
		Called at the end of a training (each minibatch) depening on the algo.
		This is also a moment to save what we have to disk.
		"""
		self.optitcall = 0
		logger.info("Cumulated training time: {0:.2f} s".format(np.sum(self.optittimes)))
		logger.info("Optimization cycle finished.")
		if self.trackbiases:
			self.savebiasdetails()
		if self.itersavepath != None:
			if not self.saveeachit: # Indeed, otherwise no need to save it at this stage!
				self.save(self.itersavepath)
		if self.autoplot:
			self.makeplots()
		

	def callback(self, *args):
		"""
		Function called by the optimizer after each "iteration".
		Print out some info about the training progress,
		saves status of the counters,
		and optionally writes the network itself to disk.
		"""
		#print args
		#exit()
		
		self.optit += 1
		now = datetime.now()
		secondstaken = (now - self.iterationstarttime).total_seconds()
		callstaken = self.optitcall 
		
		# Not sure if it is needed to update the params (of if the optimizer already did it), but it cannot harm and is fast:
		self.params[self.paramslice] = args[0] # Updates the network parameters
		
		self.optittimes.append(secondstaken)
		self.optiterrs_train.append(self.opterr)
		self.optitcalls.append(self.optcall)
		self.optitparams.append(copy.deepcopy(self.params)) # We add a copy of the current params
				
		# Now we evaluate the cost on the validation set:
		valerr = self.valcost()
		self.optiterrs_val.append(valerr)
		
		valerrratio = valerr / self.opterr
		
		mscallcase = 1000.0 * float(secondstaken) / (float(callstaken) * self.dat.getntrain()) # Time per call and training case
		
		logger.info("Iter. {self.optit:4d}, {ef} train = {self.opterr:.6e}, val = {valerr:.6e} ({valerrratio:4.1f}), {time:.4f} s for {calls} calls ({mscallcase:.4f} ms/cc)".format(
			self=self, ef=self.get_costfctname(), time=secondstaken, valerr=valerr, valerrratio=valerrratio, calls=callstaken, mscallcase=mscallcase))
		
		if self.itersavepath != None:
			if self.saveeachit:
				self.save(self.itersavepath)
		
		# We reset the iteration counters:
		self.iterationstarttime = datetime.now()
		self.optitcall = 0 
			
		# And now we take care of getting a new batch
		#self.randombatch()
		
		
		
	def cost(self, p):
		"""
		The "as-fast-as-possible" function to compute the training error based on parameters p.
		This gets called repeatedly by the optimizers.
		"""
	
		self.params[self.paramslice] = p # Updates the network parameters
		
		# Compute the outputs
		outputs = self.net.run(self.dat.traininputs) # This is not a masked array!
		
		# And now evaluate the error (cost) function.
		if self.dat.trainoutputsmask is not None:
			outputs = np.ma.array(outputs, mask=self.dat.trainoutputsmask)
		err = self.errfct(outputs, self.dat.traintargets, auxinputs=self.dat.trainauxinputs)
		
		if self.regullam is not None:
			# TODO: Is there a faster way to do this?
			weights = np.concatenate(([(l.weights).flatten() for l in self.net.layers]))
			err += self.regullam * self.regulfct(weights) 
			
		self.opterr = err
		self.optcall += 1
		self.optitcall += 1
		self.opterrs.append(err)
		
		if self.verbose:
			logger.debug("Iteration {self.optit:4d}, call number {self.optcall:8d}: cost = {self.opterr:.8e}".format(self=self))
			logger.debug("\n" + self.net.report())
			
		return err


	def currentcost(self):
		return self.cost(p=self.params[self.paramslice])


	def testcost(self):
		"""
		Calls the cost function and logs some info.
		"""
		
		logger.info("Testing cost function calls...")
		starttime = datetime.now()
		err = self.currentcost()
		endtime = datetime.now()
		took = (endtime - starttime).total_seconds()
		logger.info("On the training set:   {took:.4f} seconds, {self.errfctname} = {self.opterr:.8e}".format(self=self, took=took))
		starttime = datetime.now()
		err = self.valcost()
		endtime = datetime.now()
		took = (endtime - starttime).total_seconds()		
		logger.info("On the validation set: {took:.4f} seconds, {self.errfctname} = {err:.8e}".format(self=self, took=took, err=err))
		
	
		
	def valcost(self):
		"""
		Evaluates the cost function on the validation set.
		"""
		outputs = self.net.run(self.dat.valinputs) # This is not a masked array!
		if self.dat.valoutputsmask is not None:
			outputs = np.ma.array(outputs, mask=self.dat.valoutputsmask)
		
		err = self.errfct(outputs, self.dat.valtargets, auxinputs=self.dat.valauxinputs)
		
		return err
		
	def savebiasdetails(self):
		"""
		Appends bias details to some internal lists.
		For plotting purposes only.
		A bit ugly, all this duplication :-/
		Note that we use the FULL training data, not only the current batch.
		"""
		logger.info("Saving bias details...")
		fulltrainoutputs = self.net.run(self.dat.fulltraininputs) # This is not a masked array!
		if self.dat.fulltrainoutputsmask is not None:
			fulltrainoutputs = np.ma.array(fulltrainoutputs, mask=self.dat.fulltrainoutputsmask)		
		fulltrainbiases = np.mean(fulltrainoutputs, axis=0) - self.dat.fulltraintargets
		
		valoutputs = self.net.run(self.dat.valinputs) # This is not a masked array!
		if self.dat.valoutputsmask is not None:
			valoutputs = np.ma.array(valoutputs, mask=self.dat.valoutputsmask)
		valbiases = np.mean(valoutputs, axis=0) - self.dat.valtargets
		
		# Those 2 arrays have shape (neuron, cases)
		self.biassnaps_train.append(fulltrainbiases)
		self.biassnaps_val.append(valbiases)
		self.biassnaps_it.append(self.optit) # We record the iteration counter

		

	def opt(self, algo="brute", mbsize=None, mbfrac=0.1, mbloops=10, **kwargs):
		"""
		General optimization of the network parameters by an algorithm, with support for minibatches.
		Just set mbfrac and mbloops to 1 if you don't want minibatches.
		Using "algo" and not "method" so that "method" can be a kwarg passed to scipy.optimize.fmin...
		"""
		
		optfct = eval("opt.{0}".format(algo)) # Turn the string into an acutal function
		
		for loopi in range(mbloops):
			if mbloops > 1:
				logger.info("Starting minibatch loop {loopi} of {mbloops}...".format(loopi=loopi+1, mbloops=mbloops))
			self.dat.random_minibatch(mbsize=mbsize, mbfrac=mbfrac)
			self.optbatchchangeits.append(self.optit) # We record this minibatch change
			
			self.start()
			optfct(self, **kwargs) # Call the optfct, with the training object as first argument
			
			logger.info("Done with optimization")
			self.end()



#	
#	def minibatch_bfgs(self, mbsize=None, mbfrac=0.1, mbloops=10, **kwargs):
#		
#		for loopi in range(mbloops):
#			if mbloops > 1:
#				logger.info("Starting minibatch loop {loopi} of {mbloops}...".format(loopi=loopi+1, mbloops=mbloops))
#			self.dat.random_minibatch(mbsize=mbsize, mbfrac=mbfrac)
#			self.optbatchchangeits.append(self.optit) # We record this minibatch change
#			self.bfgs(**kwargs)
			
	

#	def bfgs(self, maxiter=100, gtol=1e-8):
#		
#		self.start()
#		logger.info("Starting BFGS for {} iterations (maximum) with gtol={}...".format(maxiter, gtol))
#		
#		optres = scipy.optimize.fmin_bfgs(
#			self.cost, self.params[self.paramslice],
#			fprime=None,
#			maxiter=maxiter, gtol=gtol,
#			full_output=True, disp=True, retall=False, callback=self.callback)
#		
#		if len(optres) == 7:
#			(xopt, fopt, gopt, Bopt, func_calls, grad_calls, warnflag) = optres
#			self.cost(xopt) # Is it important to do this, to set the optimal parameters? It seems not.
#			logger.info("Done with optimization, {0} func_calls and {1} grad_calls".format(func_calls, grad_calls))
#		else:
#			logger.warning("Optimization output is fishy")
#		
#		self.end()
	


#	def cg(self, maxiter):
#		
#		self.start()
#		logger.info("Starting CG for {0} iterations (maximum)...".format(maxiter))
#		
#		optres = scipy.optimize.fmin_cg(
#			self.cost, self.params,
#			fprime=None, gtol=1e-05,
#			maxiter=maxiter, full_output=True, disp=True, retall=False, callback=self.callback)
#			
#		if len(optres) == 5:
#			(xopt, fopt, func_calls, grad_calls, warnflag) = optres
#			self.cost(xopt) # Is it important to do this, to set the optimal parameters? It seems not.
#			logger.info("Done with optimization, {0} func_calls and {1} grad_calls".format(func_calls, grad_calls))
#		else:
#			logger.warning("Optimization output is fishy")
#	
#		self.end()


#	def anneal(self, maxiter=100):
#		
#		self.testcost()
#		logger.info("Starting annealing for {0} iterations (maximum)...".format(maxiter))
#	
#		optres = scipy.optimize.basinhopping(
#			self.cost, self.params, 
#			niter=maxiter, T=0.001, stepsize=0.1, minimizer_kwargs=None, take_step=None, accept_test=None,
#			callback=self.callback, interval=100, disp=True, niter_success=None)
#			
#			# Warning : interval is not the callback interval, but the step size update interval.
#
#		print optres
#		
#		print len(optres)
		
#	def fmin(self, maxiter=100):	# One iteration per call
#		self.testcost()
#		logger.info("Starting fmin for {0} iterations (maximum)...".format(maxiter))
#		
#		optres = scipy.optimize.fmin(
#			self.cost, self.params,
#			xtol=0.0001, ftol=0.0001, maxiter=maxiter, maxfun=None,
#			full_output=True, disp=True, retall=True, callback=self.callback)
#		
#		print optres

	
#		"""
#		optres = scipy.optimize.fmin_powell(
#			cost, params,
#			maxiter=maxiter, ftol=1e-06,
#			full_output=True, disp=True, retall=True, callback=self.optcallback)
#		"""
#		"""
#		optres = scipy.optimize.fmin(
#			cost, params,
#			xtol=0.0001, ftol=0.0001, maxiter=maxiter, maxfun=None,
#			full_output=True, disp=True, retall=True, callback=self.optcallback)
#		"""
#		"""
#		optres = scipy.optimize.minimize(
#			cost, params, method="Anneal",
#			jac=None, hess=None, hessp=None, bounds=None, constraints=(),
#			tol=None, callback=self.optcallback, options={"maxiter":maxiter, "disp":True})
#		"""
#		
#		"""
#		optres = scipy.optimize.basinhopping(
#			cost, params, 
#			niter=maxiter, T=0.001, stepsize=1.0, minimizer_kwargs=None, take_step=None, accept_test=None,
#			callback=self.optcallback, interval=50, disp=True, niter_success=None)
#		"""



"""
Create some data to play with
"""

import numpy as np
import tenbilac

import logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


nrea = 5
ncas = 500
xs = np.random.uniform(-3, 6, ncas)
ys = np.random.uniform(-10, -5, ncas)
zs = xs * ys 
inputs = np.array([xs, ys])
inputs = np.tile(inputs, (nrea, 1, 1))
inputs += 0.01*np.random.randn(inputs.size).reshape(inputs.shape)
targets = np.array([zs])


logger.info("Inputs have shape {}, targets have shape {}".format(inputs.shape, targets.shape))
tenbilac.utils.writepickle((inputs, targets), "data.pkl")








import matplotlib.pyplot as plt
import numpy as np
import tenbilac



inputs = np.linspace(-1, 1, 1000)

ws = [-2, -1.0, -0.5, 0.0, 0.5, 1.0, 2.0]
ws = [-2, -1.0, -0.5, 0.0, 0.5, 1.0, 2.0]

ws = np.linspace(-1, 1, 5)
ws = [-0.1, 0.1, 1.0, 2.0]

for w in ws:
	#plt.plot(inputs, tenbilac.act.iden(np.power(inputs, 2.0*tenbilac.act.sig(w))), label="w="+str(w))
	#plt.plot(inputs, np.real(np.power(inputs+0j,  2.0*tenbilac.act.sig(10.0*w))), label="w="+str(w))
	plt.plot(inputs, np.sign(inputs)*np.power(np.fabs(inputs),  w), label="w="+str(w))

	#plt.plot(inputs, np.real(np.power(inputs + 0j,  w)), label="w="+str(w))


plt.xlabel("input")
plt.ylabel("output = input ** w")
plt.legend()
plt.show()





"""
Helper functions related to manipulate fits images, get stamps...

"""

import os
import galsim

import numpy as np
import logging
logger = logging.getLogger(__name__)


def loadimg(imgfilepath):
	"""
	Uses GalSim to load and image from a FITS file, enforcing that the GalSim origin is (0, 0).

	:param imgfilepath: path to FITS image
	:returns: galsim image
	"""

	logger.info("Loading FITS image %s..." % (os.path.basename(imgfilepath)))
	img = galsim.fits.read(imgfilepath)
	img.setOrigin(0,0)
	logger.info("Done with loading %s, shape is %s" % (os.path.basename(imgfilepath), img.array.shape))

	#logger.warning("The origin and stampsize conventions are new and should be tested !")

	img.origimgfilepath = imgfilepath # Just to keep this somewhere

	return img


def getheader(imgfilepath, **kwargs):
	"""
	Uses GalSim to quickly load the header
	"""
	logger.info("Getting FITS header from %s..." % (os.path.basename(imgfilepath)))
	header = galsim.FitsHeader(imgfilepath, **kwargs)
	return header



def getstamp(x, y, img, stampsize, getfrom=None):
	"""
	I prepare a image stamp "centered" at position (x, y) of your input image.
	You can use the array attribute of the stamp if you want to get the actual pixels.
	
	:param x: x position of the stamp "center" (i.e., the object), in pixels.
	:param y: 
	:param img: input numpy or galsim image. Automatically type detection.
	:param stampsize: width = height of the stamp, in pixels. Has to be even. 
	:param getfrom: force the detection "gs" for galsim and "np" for numpy.
	
	:returns: a tuple(stamp, flag). Flag is 1 if the stamp could not be extracted, 0 otherwise.
	"""
	if getfrom==None:
		if type(img)==galsim.image.Image:
			logger.debug("Found image type galsim")
			# Nice but way too verbatim
			getfrom="gs"
		elif type(img)==np.ndarray:
			logger.debug("Found image type numpy")
			# Nice but way too verbatim
			getfrom="np"
		else:
			raise TypeError("Image type not understood.")
	else:
		logger.debug("getfrom is forced to %s" % getfrom)
		
	if not int(stampsize)%2 == 0: raise ValueError("stampsize must be even.")
	
	error=False
	
	# If the stamp is from galsim:
	if getfrom=="gs":
		assert img.origin().x == 0 and img.origin().y == 0
		assert img.xmin == 0 and img.ymin == 0
		
		xmin = int(np.round(x - 0.5)) - int(stampsize)/2
		xmax = int(np.round(x - 0.5)) + int(stampsize)/2 - 1
		ymin = int(np.round(y - 0.5)) - int(stampsize)/2
		ymax = int(np.round(y - 0.5)) + int(stampsize)/2 - 1
				
		assert ymax - ymin == stampsize - 1 # This is the GalSim convention, both extermas are "included" in the bounds.
		assert xmax - xmin == stampsize - 1
		
		# We check that these bounds are fully within the image
		if xmin < img.getXMin() or xmax > img.getXMax() or ymin < img.getYMin() or ymax > img.getYMax():
			error=True
		else:			
			# We prepare the stamp
			bounds = galsim.BoundsI(xmin, xmax, ymin, ymax)
			stamp = img[bounds] # galaxy postage stamp
			assert stamp.array.shape == (stampsize, stampsize)
	# if from numpy:
	elif getfrom=="np":
		# By MomentsML's definition, a pixel is centered at 0.5,0.5
		dd=+.5
		xmin=int(round(x-dd-stampsize/2.))
		xmax=int(round(x-dd+stampsize/2.))
		ymin=int(round(y-dd-stampsize/2.))
		ymax=int(round(y-dd+stampsize/2.))
		
		# We check that these bounds are fully within the image
		if xmin < 0 or xmax > np.shape(img)[0] or ymin < 0 or ymax > np.shape(img)[1]:
			error=True
		else:
			# We prepare the stamp
			stamp=img[xmin:xmax,ymin:ymax]
			assert np.shape(stamp) == (stampsize, stampsize)
	
	if error:
		return (None, 1)
	else: 
		return (stamp, 0)



"""
Top-level classes and functions to help running MomentsML on GREAT3.
"""

import logging
logger = logging.getLogger(__name__)

from astropy.table import Table
import os

import utils
from momentsml import tools



class GREAT3Run(utils.Branch):
	"""
	This is a simple class to group frequently used variables, on top of the Branch class.
	Unlike Branch, it does specify paths to MomentsML-internal temporary files and directories, and handles a workdir.
	"""
	
	def __init__(self, experiment, obstype, sheartype, datadir, truthdir, workdir, g3publicdir, subfields=None, ncpu=None, skipdone=False):
		
		utils.Branch.__init__(self, experiment, obstype, sheartype, datadir, truthdir)
		logger.info("Getting ready to work on branch %s-%s-%s" % (experiment, obstype, sheartype))

		self.workdir=workdir
		if self.workdir == None:
			logger.warning("Better specify a workdir, I think.")
			self.workdir = "./%s" % (self.get_branchacronym())
		self.mkdirs()
		
		self.g3publicdir = g3publicdir
		
		self.subfields=subfields
		if self.subfields is None:
			self.subfields=range(200)
			
		self.ncpu = ncpu
		if ncpu is None:
			self.ncpu = 1
			
		self.skipdone = skipdone

		# Those, and further variables, can be wildly added later:
		self.simparams_name = None
		self.trainparams_name = None
		

	
	def __str__(self):
		"""
		A tiny self-description, for logging
		"""
		return "GREAT3Run on branch %s in workdir '%s'" % (self.get_branchacronym(), self.workdir)

		
	def mkdirs(self, subfield=None):
		"""
		Creates the working directories. 
		"""

		if not os.path.isdir(self.workdir):
			os.makedirs(self.workdir)
	
		if subfield is not None:
			dirpath = self.subpath(subfield)
			if not os.path.isdir(dirpath):
				os.makedirs(dirpath)
				
			# Now must create the sub-directories:
			for subfolder in ["obs","sim","ml","pred","out","val"]:
				dirpath = self.subpath(subfield, subfolder)
				if not os.path.isdir(dirpath):
					os.makedirs(dirpath)


	def path(self,*args):
		"""
		A helper function that returns a filepath within the working directory.
		
		:param args: strings, must be in order of the filepath, similar to os.path.join()
		
		Example usage::
		
			>>> self.path("obs","catalogue_000.fits")
			
		will return the filepath: self.workdir/obs/catalogue_000.fits
		"""
		return os.path.join(self.workdir,"/".join(args))
	
	
	def subpath(self, subfield, *args):
		"""
		Similar, but first argument is a subfield number
		"""
		
		
		
		return os.path.join(self.workdir, "%03i" % subfield, "/".join(args))
	




import config
import measfcts
import os
import numpy as np
import matplotlib.ticker as ticker
import matplotlib

import momentsml.plot
from momentsml.tools.feature import Feature
import matplotlib.pyplot as plt
plt.rc('text', usetex=True)

import logging
logger = logging.getLogger(__name__)

#momentsml.plot.figures.set_fancy(14)
from matplotlib import rc
rc('font',**{'family':'sans-serif','sans-serif':['Helvetica']})
## for Palatino and other serif fonts use:
#rc('font',**{'family':'serif','serif':['Palatino']})
rc('text', usetex=True)



valcat = os.path.join(config.valdir, config.valname + ".pkl")
cat = momentsml.tools.io.readpickle(valcat)

#print momentsml.tools.table.info(cat)



cat = momentsml.tools.table.shuffle(cat) # otherwise scatter plots look weird as they got sorted by tru s1 s2

momentsml.tools.table.addstats(cat, "snr")


#select = False
#if select:
#	
#	s = momentsml.tools.table.Selector("snr_mean > 10", [
#		("min", "snr_mean", 10.0),
#	])
#	catsel = s.select(cat)


#tru_sb = Feature("tru_sb", 0, 16, nicename=r"Surface brightness $S$ [pix$^{-2}$]")
#tru_rad = Feature("tru_rad", 1.5, 8.5, nicename=r"Half-light radius $R$ [pix]")

tru_mag = Feature("tru_mag", nicename=r"Magnitude")
tru_rad = Feature("tru_rad", nicename=r"Half-light radius $R$ [pix]")

snr_mean = Feature("snr_mean", nicename="$\\langle \mathrm{S}/\mathrm{N}\\rangle$")

#for comp in ["1","2"]:
for comp in ["1"]:

	
	momentsml.tools.table.addrmsd(cat, "pre_s{}".format(comp), "tru_s{}".format(comp))
	momentsml.tools.table.addstats(cat, "pre_s{}".format(comp))
	
	cat["abs_pre_s{}_bias".format(comp)] = np.fabs(cat["pre_s{}_bias".format(comp)])
	cat["log_abs_pre_s{}_bias".format(comp)] = np.log10(np.fabs(cat["pre_s{}_bias".format(comp)]))
	
	
	s_high = momentsml.tools.table.Selector("snr_mean > 10", [
		("min", "snr_mean", 10.0),
	])
	s_low = momentsml.tools.table.Selector("snr_mean < 10", [
		("max", "snr_mean", 10.0),
	])
	
	cat_high = s_high.select(cat)
	cat_low = s_low.select(cat)


	max_bias = np.max(cat["abs_pre_s{}_bias".format(comp)])
	hard_coded_max_bias = max_bias # 0.02
	max_bias_high = np.max(cat_high["abs_pre_s{}_bias".format(comp)])
	hard_coded_max_bias_high = 0.008
	max_bias_low = np.max(cat_low["abs_pre_s{}_bias".format(comp)])
	hard_coded_max_bias_low = max_bias_low # 0.02
	
	# latex and format are ugly to mix, doing it manually:
	if comp == "1":
		tru_s = Feature("tru_s{}".format(comp), -0.12, 0.12, nicename=r"$g_1^{\mathrm{true}}$")
		pre_s_mean = Feature("pre_s{}_mean".format(comp), nicename=r"Estimated $\hat{g}_{1}$ (average over realizations)")
		pre_s_bias = Feature("pre_s{}_bias".format(comp), -max_bias, max_bias, nicename=r"$\langle \hat{g}_{1} \rangle - g_{1}^{\mathrm{true}} $")
		pre_s_bias_fix = Feature("pre_s{}_bias".format(comp), -hard_coded_max_bias_high, hard_coded_max_bias_high, nicename=r"$\langle \hat{g}_{1} \rangle - g_{1}^{\mathrm{true}} $")
		abs_pre_s_bias = Feature("abs_pre_s{}_bias".format(comp), 1e-4, 1e-1, nicename=r"$|\langle \hat{g}_{1} \rangle - g_{1}^{\mathrm{true}}|$ ")
		#log_abs_pre_s_bias = Feature("log_abs_pre_s{}_bias".format(comp), nicename=r"$\log(|g_1|$ estimation error)")
		
	if comp == "2":
		tru_s = Feature("tru_s{}".format(comp), -0.12, 0.12, nicename=r"$g_2^{\mathrm{true}}$")
		pre_s_mean = Feature("pre_s{}_mean".format(comp), nicename=r"Estimated $\hat{g}_{2}$ (average over realizations)")
		pre_s_bias = Feature("pre_s{}_bias".format(comp), -max_bias_high, max_bias_high, nicename=r"$\langle \hat{g}_{2} \rangle - g_{2}^{\mathrm{true}} $")
		pre_s_bias_fix = Feature("pre_s{}_bias".format(comp), -hard_coded_max_bias_high, hard_coded_max_bias_high, nicename=r"$\langle \hat{g}_{2} \rangle - g_{2}^{\mathrm{true}} $")
		abs_pre_s_bias = Feature("abs_pre_s{}_bias".format(comp), 1e-4, 1e-1, nicename=r"$|\langle \hat{g}_{2} \rangle - g_{2}^{\mathrm{true}} |$ ")
		#log_abs_pre_s_bias = Feature("log_abs_pre_s{}_bias".format(comp), nicename=r"$\log(|g_2|$ estimation error)")
	
		


	#fig = plt.figure(figsize=(11.5, 3.0))
	fig = plt.figure(figsize=(9.0, 6.0))
	plt.subplots_adjust(
		left  = 0.12,  # the left side of the subplots of the figure
		right = 0.9,    # the right side of the subplots of the figure
		bottom = 0.1,   # the bottom of the subplots of the figure
		top = 0.95,      # the top of the subplots of the figure
		wspace = 0.7,   # the amount of width reserved for blank space between subplots,
	                # expressed as a fraction of the average axis width
		hspace = 0.3,   # the amount of height reserved for white space between subplots,
					# expressed as a fraction of the average axis heightbottom=0.1, right=0.8, top=0.9)
		)


	rasterized=True

	ax = fig.add_subplot(2, 2, 1)
	momentsml.plot.scatter.scatter(ax, cat_high, tru_s,  pre_s_bias, featc=snr_mean, cmap="plasma_r", s=10, rasterized=rasterized)
	ax.set_ylim((-0.01, 0.01))
	ax.text(0.1, 0.87, r"$\langle \mathrm{S}/\mathrm{N} \rangle > 10$", verticalalignment='center', transform=ax.transAxes)
	ax.axhline(color="black", lw=0.5)
	
	ax = fig.add_subplot(2, 2, 2)
	momentsml.plot.scatter.scatter(ax, cat_low, tru_s,  pre_s_bias, featc=snr_mean, cmap="plasma_r", s=10, rasterized=rasterized)
	ax.set_ylim((-0.1, 0.1))
	ax.text(0.1, 0.87, r"$\langle \mathrm{S}/\mathrm{N} \rangle < 10$", verticalalignment='center', transform=ax.transAxes)
	ax.axhline(color="black", lw=0.5)
	
	
	ax = fig.add_subplot(2, 2, 3)
	cnorm = matplotlib.colors.SymLogNorm(linthresh=0.005)
	momentsml.plot.scatter.scatter(ax, cat_high, tru_s, tru_rad, featc=pre_s_bias_fix, cmap="coolwarm", norm=cnorm, s=10, rasterized=rasterized)
	ax.text(0.1, 0.87, r"$\langle \mathrm{S}/\mathrm{N} \rangle > 10$", verticalalignment='center', transform=ax.transAxes)
	
	
	"""
	ax = fig.add_subplot(1, 3, 3)
	cnorm = matplotlib.colors.SymLogNorm(linthresh=0.001)
	momentsml.plot.scatter.scatter(ax, cat, tru_sb, tru_rad, featc=pre_s_bias, cmap="coolwarm")
	"""
	
	ax = fig.add_subplot(2, 2, 4)
	cnorm = matplotlib.colors.LogNorm(vmin=1e-4, vmax=1e-1, clip=True)
	momentsml.plot.scatter.scatter(ax, cat, tru_mag, tru_rad, featc=abs_pre_s_bias, cmap="plasma_r", s=10, norm=cnorm, rasterized=rasterized)
	
	# Now we plot the snr limit
	
	
	#plt.tight_layout()


	momentsml.plot.figures.savefig(os.path.join(config.valdir, config.valname + "_resi_{}".format(comp)), fig, fancy=True, pdf_transparence=True)
	plt.show()




"""
Helpers to specify a column of an astropy table.
"""

import numpy as np
import astropy.table

import logging
logger = logging.getLogger(__name__)


class Feature():
	"""
	This class holds a column name together with an interval to nicely show 
	this column values on plots, and also to select a given "realization" of a column.
	It's called "Feature", but it can well apply to any column of an astropy table (including "labels", in machine learning speak...).
	One can also specify a column name for the uncertainties (that will typically be shown as 1-sigma error bars).
	It is not tied to a particular table, nor to a particular plot: the same "Feature" object can be reused for different tables and plots.
	"""
	def __init__(self, colname, low=None, high=None, nicename=None, errcolname=None, rea=None):
		"""

		:param colanme: name of a catalog column containing the values to be plotted 
		:param low: (optional) lower bound of the interval to show on the plot
		:param high: (optional) higher bound
		:param nicename: (optional) a "nice" string designating the feature, to be used as axis label.
		:param errcolname: (optional) name of a column containing uncertainties / errorbars
		:param rea: specifies which realization to use, if several are available. "full" means all. 0 means first (etc), -5 means "the first 5 reas".
		:type rea: int or string
		
		"""
		
		self.colname = colname
		self.errcolname = errcolname
		self.low = low
		self.high = high
		
		self.rea = rea
		
		if nicename == None:
			self.nicename = self.colname
		else:
			self.nicename = nicename


	def err(self):
		"""
		Returns a new Feature describing the errcolname, if this was specified.
		This is useful to get one Feature per column that is required for a plot.
		"""
		if self.errcolname is not None:
			return Feature(self.errcolname, rea=self.rea)
		else:
			return None


	def get(self, cat, flatten=False):
		"""
		Returns the column corresponding to the feature, from the given catalog (as masked, if the column was masked).
		We don't do anything about the mask here.
		The point here is to get nice automatic behaviour even for 2D columns.
		If flatten is True, 2D columns are returned as flattened to 1D.
		Note that we use the default "C" order for this flattening. One could use "F", so that the output starts with the first realization of all the rows, then the second realization, etc.
		But any "inconsistency" in the catalog would already be done before. Switching to F only helps if we also change the way meas.avg.onsims stacks its stuff.
		
		"""
		
		
		if self.colname not in cat.colnames:
			raise RuntimeError("The column '%s' is not available in %s" % (self.colname, cat.colnames))
				
		if cat[self.colname].ndim == 1:
			# Things are easy, we don't need rea.
			# Let's check if there is a rea, this could be a mistake.
			if not self.rea is None:
				logger.warning("Feature {self.colname} has rea={self.rea} but the column is 1D!".format(self=self))
			output = cat[self.colname]
		
		elif cat[self.colname].ndim == 2:
		
			
			if cat[self.colname].shape[1] == 1: # Then in fact it's a 1D column, with just one value per cell...
				# We simply return the data from this colum, no need to care for rea.
				output = cat[self.colname] #.flatten() # Not sure if would be logic to flatten it here. No, why would you do that ?
			
			elif self.rea is None:
				raise RuntimeError("The column {self.colname} is 2D, please specify a rea for this feature!".format(self=self))
			
			elif type(self.rea) == int:
				
				if self.rea >= 0:
					output = cat[self.colname][:,self.rea]
					assert output.ndim == 1
					logger.info("Only using the particular rea {self.rea} of 2D colum '{self.colname}'".format(self=self))
					
				else: # Negative rea, means: use the first -rea realizations!
					
					#if -self.rea > cat[self.colname].shape[1]
					#	raise RuntimeError("Cannot select rea = {0} from a column with shape {1}".format(self.rea, cat[self.colname].shape))
				
					output = cat[self.colname][:,0:-self.rea]
					logger.info("Using the first {nrea}/{totrea} reas of 2D colum '{self.colname}'".format(self=self, nrea=-self.rea, totrea=cat[self.colname].shape[1]))
				
			elif self.rea in ["full", "Full", "All", "all"]:
				
				output = cat[self.colname]
				assert output.ndim == 2
				logger.info("Using the full {nrea} reas of 2D colum '{self.colname}'".format(self=self, nrea=cat[self.colname].shape[1]))
				
			else:
				raise RuntimeError("Could not understand rea={self.rea} of the feature {self.colname}.".format(self=self))
			
		else:
			raise RuntimeError("Weird column shape!")
		
		if output.ndim != 1 and flatten:
			return output.T.flatten()	
		else:
			return output
				

def get1Ddata(cat, features, keepmasked=True):
	"""
	Prepares 1D arrays of the same length, one for each feature.
	If features contains both 2D and 1D stuff, this is done by reapeating the 1D stuff. In other words, it intelligently flattens realizations.
	We don't do anything special about masks, they get carried arround as expected. Except if keepmasked is False, than we keep only unmasked points at the end.
	This is the function that you want to use when preparing plots or computing metrics...
	"""
	
	# First, we check the features, to see if the user wants to mix different realizations:
	reasettings = [feature.rea for feature in features if feature.rea != None]
	reasettings = list(set(reasettings))
	if len(reasettings) > 1:
		logger.warning("I was asked to identify data from different realizations, this is probably a mistake!")
	
	catlen = len(cat)
	data = []
	for f in features:
		d = f.get(cat, flatten=False) # We do NOT flatten the columns
		assert d.ndim <= 2 # Columns shound not be 3D...
		assert len(d) == catlen # All columns have the same length, only the 2nd dimension can change.
		
		if d.ndim == 2 and d.shape[1] == 1: # Then it's a 2D column but containing only one "realization".
			# We flatten it to make a plain 1D column, just for consistency.
			d = d.flatten()
			assert d.ndim == 1
		
		data.append(d)
	
	assert len(data) == len(features)

	# Now we check the dimensions of these columns.
	
	nbreas = [] # Number of realizations for each column.
	for d in data:
		if d.ndim == 1:
			nbreas.append(1)
		elif d.ndim == 2:
			nbreas.append(d.shape[1])
			
	assert len(nbreas) == len(data)
	
	unilengths = list(set(nbreas))
	
	if len(unilengths) > 2:
		raise RuntimeError("Mixture of number of realizations is weird or too complicated: {0}".format(unilengths))
	
	elif len(unilengths) == 1:
		if unilengths[0] == 1: # Then we only have 1D columns, and we are done.
		
			data1d = data
		
		else: # Then all columns are 2D but have the same number of realizations, very easy:
			
			data1d = [d.flatten() for d in data]
	
	elif len(unilengths) == 2: # Then we have a mixture of different number of realizations, and need to "tile".
		small = min(unilengths)
		large = max(unilengths)
		if small != 1:
			raise RuntimeError("For safety reasons I don't want to use tiling to mix columns with different numbers of realizations, even if I could.")
		
		data1d = []
		for (d, nbrea) in zip(data, nbreas):
			if nbrea == 1: # We have a 1D column
				
				# We have to "tile" it.
				# We do this by indeed tiling it to 2D, and then flattenting it in the same way as the 2D cols.
				# This seems "safest", as we have to be very careful here not to mess up with the flatten ordering !
				
				tiledd = np.tile(d, (large, 1)).T
				assert tiledd.ndim == 2
				assert tiledd.shape[0] == catlen
				assert tiledd.shape[1] == large
				data1d.append(tiledd.flatten())
				
			elif nbrea == large:
				data1d.append(d.flatten())
				
			else:
				raise RuntimeError("Does not happen.")
		
	assert len(data1d) == len(data)
	
	# We return this as a table.
	colnames = [f.colname for f in features]
	table1d = astropy.table.Table(data1d, names=colnames)
	
	if keepmasked == True:
		return table1d
		
	else:
		# We prepare a version of "features" with all reas set to 1D, so that we can use cutmasked:
		features1d = [Feature(colname=f.colname, rea=None) for f in features]
		return cutmasked(table1d, features1d)

	
#	An old version of it:	
# 
#	data = []
#	for f in features:
#		d = f.get(cat, flatten=True)
#		assert d.ndim == 1
#		data.append(d)
#		
#	lengths = [d.size for d in data]
#	unilengths = list(set(lengths)) # Unique list of lenghts.
#	
#	
#	fixeddata = []
#	
#	if len(unilengths) == 1: # Nothing special to do, then
#		fixeddata = data	
#
#	elif len(unilengths) == 2:
#	
#		small = min(unilengths)
#		large = max(unilengths)
#		(div, mod) = divmod(large, small)
#			
#		if mod != 0:
#			raise RuntimeError("Cannot work with lengths %i and %i" % (large, small))
#		
#		logger.warning("Mild warning: some of the columns need to be tiled {0} times, and at this stage we cannot verify if the ordering of this tiling is as intended.".format(div))
#		
#		# We repeat the smaller column div times...
#		for d in data:
#			if d.size == small:
#				fixedd = np.tile(d, div)
#				assert fixedd.size == large
#				fixeddata.append(fixedd)
#			else:
#				fixeddata.append(d)
#	else:
#		raise RuntimeError("Mixture is weird or too complicated: {0}".format(unilengths))
#		
#	assert len(fixeddata) == len(features)
#	
#	# We return this as a table.
#	colnames = [f.colname for f in features]
#	
#	table1d = astropy.table.Table(fixeddata, names=colnames)
#	
#	
#	if keepmasked == True:
#		return table1d
#		
#	else:
#		# We prepare a version of "features" with all reas set to 1D, so that we can use cutmasked:
#		features1d = [Feature(colname=f.colname, rea=None) for f in features]
#		return cutmasked(table1d, features1d)
	
	
	


def cutmasked(cat, features):
	"""
	Prepares a copy of cat containing only those rows for which none of the features is masked.
	To do this, it cleverly combines the masks.
	This is in principle close to what **numpy.ma.compress_rows** does.
	But here we also log a bit about the masked columns, and we take care of 2D columns by respecting the "rea" of Feature.

	
	:param cat: an astropy table
	:param features: a list of Feature objects
	
	:returns: a table containing only the available rows 
	
	"""
	
	if len(cat) == 0:
		raise RuntimeError("Give me a non-empty catalog.")
	# The first objective is to get a list of 1D masks, one per column.
	
	masklist = []
	colnames = []
	
	for f in features:
	
		column = f.get(cat)
		mask = np.array(np.ma.getmaskarray(column), dtype=bool)
		assert mask.ndim in [1, 2]
		if mask.ndim == 2: # This means if rea was "full"
			mask = np.any(mask, axis=1)
		assert mask.ndim == 1
		assert mask.size == len(cat)
		masklist.append(mask)
		colnames.append(f.colname)	
		
	
	# We now group all the masks into a single 2D array.
	assert len(colnames) == len(masklist)
	masks = np.column_stack(masklist)
	
	
	# Now we log some details about how many points are masked in each column.
	for (i, colname) in enumerate(colnames):
		mask = masks[:,i]
		assert len(mask) == len(cat)
		nbad = np.sum(mask)
		logger.info("Column %20s: %5i (%5.2f %%) of the %i entries are masked" \
			% ("'"+colname+"'", nbad, 100.0 * float(nbad) / float(len(cat)), len(cat)))
	
	# Now we combine the masks:
	combimask = np.logical_not(np.sum(masks, axis=1).astype(bool)) # So "True" means "keep this".
	assert combimask.size == len(cat)
	ngood = np.sum(combimask)
	nbad = combimask.size - ngood
	logger.info("Combination: %i out of %i (%.2f %%) rows have masked values and will be disregarded" \
		% (nbad, combimask.size, 100.0 * float(nbad) / float(combimask.size)))
	
	# We disregard the masked rows:
	nomaskcat = cat[combimask] # Only works as combimask is a numpy array ! It would do something else with a list !
	# Wow, this is extremely SLOW !
	# Note that everything so far also works for columns which are **not** masked. They will remain maskless.
	
	# More asserts:
	assert len(nomaskcat) == ngood
	#for colname in colnames:
	#	assert np.any(np.ma.getmaskarray(nomaskcat[colname])) == False # No item should be masked
	# No, as some other reas might be masked.	
	
	
	return nomaskcat


	
#	
#	"""
#	# Potentially, these columns can exist:	
#	colnames = []
#	for feature in features:
#		colnames.append(feature.colname)
#		if feature.errcolname != None:
#			colnames.append(feature.errcolname)
#	
#	colnames = list(set(colnames)) # To ensure that they are unique
#	
#	# Some trivial tests:
#	for colname in colnames:
#		if colname not in cat.colnames:
#			raise RuntimeError("The column '%s' is not available among %s" % (colname, str(cat.colnames)))
#	
#	"""
#	"""
#		#assert cat[colname].ndim == 1 # This function has not been tested for anything else...
#		# No, this is now beginnign to get implemented...
#			
#	if len(colnames) != len(list(set(colnames))):
#		raise RuntimeError("Strange, some colnames appear multiple times in %s" % (str(colnames)))
#	
#	# First, get a list of 1D masks, one per column
#	"""
#	
#	masklist = []
#	
#	print "Terrible hack, just uses mask of first rea if 2D !"
#	
#	for colname in colnames:
#		
#		mask = np.array(np.ma.getmaskarray(np.ma.array(cat[colname])), dtype=bool)
#		assert mask.ndim in [1, 2]
#		if mask.ndim == 2:
#			mask = mask[:,0] # Just get the first realization..
#			
#		assert mask.ndim == 1
#		assert mask.size == len(cat)
#		masklist.append(mask)
#	
#	# We now group all the masks for all these columns.
#	masks = np.column_stack(masklist)
#	
#	# This is a 2D boolean array. Now we log some details about how many points are masked in each column.
#	for (i, colname) in enumerate(colnames):
#		mask = masks[:,i]
#		assert len(mask) == len(cat)
#		nbad = np.sum(mask)
#		logger.info("Column %20s: %5i (%5.2f %%) of the %i entries are masked" \
#			% ("'"+colname+"'", nbad, 100.0 * float(nbad) / float(len(cat)), len(cat)))
#	
#	# Now we combine the masks:
#	combimask = np.logical_not(np.sum(masks, axis=1).astype(bool)) # So "True" means "keep this".
#	assert combimask.size == len(cat)
#	ngood = np.sum(combimask)
#	nbad = combimask.size - ngood
#	logger.info("Combination: %i out of %i (%.2f %%) rows have masked values and will be disregarded" \
#		% (nbad, combimask.size, 100.0 * float(nbad) / float(combimask.size)))
#	
#	# We disregard the masked rows:
#	nomaskcat = cat[combimask] # Only works as combimask is a numpy array ! It would do something else with a list !
#	# Wow, this is extremely SLOW !
#	# Note that everything so far also works for columns which are **not** masked. They will remain maskless.
#	
#	# More asserts:
#	assert len(nomaskcat) == ngood
#	for colname in colnames:
#		if hasattr(nomaskcat[colname], "mask"): # We make this test only if the column is masked
#			#assert np.all(np.logical_not(nomaskcat[colname].mask)) == True # Tests that all values are unmasked.
#			pass
#	
#	if not keep_all_columns:
#		nomaskcat.keep_columns(colnames)
#		# Moving this to before the slow "cat[combimask]" seems to not speedup this function -- strange ?
#	
#	return nomaskcat
#



import momentsml
import os
import config
import numpy as np

from momentsml.tools.feature import Feature

import matplotlib
import matplotlib.pyplot as plt

#momentsml.plot.figures.set_fancy(14)
from matplotlib import rc
rc('font',**{'family':'sans-serif','sans-serif':['Helvetica']})
## for Palatino and other serif fonts use:
#rc('font',**{'family':'serif','serif':['Palatino']})
rc('text', usetex=True)


cat = momentsml.tools.io.readpickle(os.path.join(config.workdir, "cat.pkl"))

print momentsml.tools.table.info(cat)


fig = plt.figure(figsize=(4, 2.5))
plt.subplots_adjust(
	left  = 0.15,  # the left side of the subplots of the figure
	right = 0.95,    # the right side of the subplots of the figure
	bottom = 0.2,   # the bottom of the subplots of the figure
	top = 0.95,      # the top of the subplots of the figure
	wspace = 0.1,   # the amount of width reserved for blank space between subplots,
	                # expressed as a fraction of the average axis width
	hspace = 0.25,   # the amount of height reserved for white space between subplots,
					# expressed as a fraction of the average axis heightbottom=0.1, right=0.8, top=0.9)
	)


tru_rad = Feature("tru_rad", -1, 13, nicename="Half-light radius [VIS pix]")
tru_mag = Feature("tru_mag", 20, 25.5, nicename="Magnitude")
tru_sersicn = Feature("tru_sersicn", 0, 6.5, nicename=r"${\textrm{S\'ersic}$ index")



ax = fig.add_subplot(1, 1, 1)
momentsml.plot.hist.hist(ax, cat, tru_sersicn, normed=True)
ax.set_ylabel("Relative frequency")
#ax.set_yticklabels([])


momentsml.plot.figures.savefig(os.path.join(config.valdir, "sourcecat2"), fig, fancy=True, pdf_transparence=True, nocrop=True)

#plt.tight_layout()
plt.show()
plt.close(fig) # Helps releasing memory when calling in large loops.






import os
import simparamsgauss
import momentsml
import momentsml.meas
import numpy as np

import logging

#logging.basicConfig(level=logging.INFO)
logging.basicConfig(format='PID %(process)06d | %(asctime)s | %(levelname)s: %(name)s(%(funcName)s): %(message)s',level=logging.INFO)



#### Parameters ###################

genworkdir = "/vol/fohlen11/fohlen11_1/mtewes/snr_fig/"
sexpath="/vol/software/software/astro/sextractor/sextractor-2.19.5/64bit/bin/sex"
stampsize = 48

name = "v1"

n = 12
nc = 3
nrea = 100
ncpu = 10
stampsize = 30
gain = 1.0e12

sexpath="/vol/software/software/astro/sextractor/sextractor-2.19.5/64bit/bin/sex"

####################################



sp = simparamsgauss.Simple1()


workdir = os.path.join(genworkdir, name)
if not os.path.isdir(workdir):
	os.makedirs(workdir)
fitsimgpath = os.path.join(workdir, "img.fits")
catpath = os.path.join(workdir, "cat.pkl")
writecatpath = os.path.join(workdir, "writecat.txt")


cat = momentsml.sim.stampgrid.drawcat(sp, n=n, nc=nc, stampsize=stampsize)

momentsml.sim.stampgrid.drawimg(cat, simgalimgfilepath=fitsimgpath, simtrugalimgfilepath=None, simpsfimgfilepath=None)


def measfct(catalog, stampsize):
	"""
	Default measfct, runs on "img".
	"""	
	# HSM adamom
	catalog = momentsml.meas.galsim_adamom.measfct(catalog, stampsize=stampsize, variant="wider")
	catalog = momentsml.meas.adamom_calc.measfct(catalog)
	# And skystats
	catalog = momentsml.meas.skystats.measfct(catalog, stampsize=stampsize)
	# And snr
	catalog = momentsml.meas.snr.measfct(catalog, gain=gain) # Gain set to give sky-limited SNR
	
	# And SExtractor
	params = ["VECTOR_ASSOC(3)", "XWIN_IMAGE", "YWIN_IMAGE", "AWIN_IMAGE", "BWIN_IMAGE", "THETAWIN_IMAGE",
	"FLUX_WIN", "FLUXERR_WIN", "NITER_WIN", "FLAGS_WIN", "FLUX_AUTO", "FLUXERR_AUTO",
	"FWHM_IMAGE", "BACKGROUND", "FLAGS",
	"FLUX_ISO", "FLUXERR_ISO"]
	
	config = {"DETECT_MINAREA":5, "ASSOC_RADIUS":5, "GAIN":gain, "ASSOC_TYPE":"NEAREST"}
		
	# FOR SOME STRANGE REASON, I CAN'T GET SEXTRACTOR TO WORK HERE
	#catalog = momentsml.meas.sewfunc.measfct(catalog, params=params, config=config, sexpath=sexpath)

	#catalog["sex_snr_iso"] = catalog["sewpy_FLUX_ISO"] / catalog["sewpy_FLUXERR_ISO"]
	#catalog["sex_snr_auto"] = catalog["sewpy_FLUX_AUTO"] / catalog["sewpy_FLUXERR_AUTO"]

	return catalog
	

groupcols = [
'adamom_flag',
'adamom_flux',
'adamom_x',
'adamom_y',
'adamom_g1',
'adamom_g2',
'adamom_sigma',
'adamom_rho4',
'adamom_logflux',
'adamom_g',
'adamom_theta',
'skystd',
'skymad',
'skymean',
'skymed',
'skystampsum',
'skyflag',
'snr'
#'sewpy_XWIN_IMAGE',
#'sewpy_YWIN_IMAGE',
#'sewpy_AWIN_IMAGE',
#'sewpy_BWIN_IMAGE',
#'sewpy_THETAWIN_IMAGE',
#'sewpy_FLUX_WIN',
#'sewpy_FLUXERR_WIN',
#'sewpy_NITER_WIN',
#'sewpy_FLAGS_WIN',
#'sewpy_FLUX_AUTO',
#'sewpy_FLUXERR_AUTO',
#'sewpy_FWHM_IMAGE',
#'sewpy_KRON_RADIUS',
#'sewpy_FLUX_RADIUS',
#'sewpy_FLUX_RADIUS_1',
#'sewpy_FLUX_RADIUS_2',
#'sewpy_FLUX_RADIUS_3',
#'sewpy_FLUX_RADIUS_4',
#'sewpy_FLUX_RADIUS_5',
#'sewpy_FLUX_RADIUS_6',
#'sewpy_FLUX_APER',
#'sewpy_FLUX_APER_1',
#'sewpy_FLUXERR_APER',
#'sewpy_FLUXERR_APER_1',
#'sewpy_BACKGROUND',
#'sewpy_FLAGS',
#'sewpy_assoc_flag',
#'sewpy_snr',
]
removecols = []

momentsml.sim.run.multi(
	simdir=workdir,
	simparams=sp,
	drawcatkwargs={"n":n, "nc":nc, "stampsize":stampsize},
	drawimgkwargs={}, 
	psfcat=None, psfselect="random",
	ncat=1, nrea=nrea, ncpu=ncpu,
	savepsfimg=False, savetrugalimg=False
)


# Measuring the newly drawn images
momentsml.meas.run.onsims(
	simdir=workdir,
	simparams=sp,
	measdir=workdir,
	measfct=measfct,
	measfctkwargs={"stampsize":stampsize},
	ncpu=ncpu,
	skipdone=False
)
	

cat = momentsml.meas.avg.onsims(
	measdir=workdir, 
	simparams=sp,
	task="group",
	groupcols=groupcols, 
	removecols=removecols
)



"""
cat.meta["img"] = momentsml.tools.imageinfo.ImageInfo(
	fitsimgpath,
	xname="x",
	yname="y",
	stampsize=stampsize,
	pixelscale=1.0
	)

cat = momentsml.meas.galsim_adamom.measfct(cat, stampsize=stampsize, variant="wider")
cat = momentsml.meas.skystats.measfct(cat, stampsize=stampsize)
#cat = momentsml.meas.snr.measfct(cat, gain=1.0)
cat = momentsml.meas.snr.measfct(cat, gain=gain)
"""


cat = momentsml.tools.table.groupreshape(cat, groupcolnames=["tru_flux", "tru_sigma", "x", "y"])
momentsml.tools.table.addstats(cat, "snr")
#momentsml.tools.table.addstats(cat, "sex_snr_iso")
#momentsml.tools.table.addstats(cat, "sex_snr_auto")
momentsml.tools.table.addstats(cat, "adamom_sigma")
momentsml.tools.table.addstats(cat, "adamom_flux")

momentsml.tools.io.writepickle(cat, catpath)

cat.sort(["y", "x"])

print momentsml.tools.table.info(cat)
#writecat = cat["tru_x", "tru_y", "tru_flux", "tru_sigma", "snr_mean", "sex_snr_iso_mean", "sex_snr_auto_mean", "adamom_sigma_mean", "adamom_flux_mean"]
writecat = cat["x", "y", "snr_n", "tru_flux", "adamom_flux_mean", "tru_sigma", "adamom_sigma_mean", "snr_mean"]

print writecat
writecat.write(writecatpath, format="ascii")





import momentsml
import os
import config
import numpy as np

from momentsml.tools.feature import Feature

import matplotlib
import matplotlib.pyplot as plt



cat = momentsml.tools.io.readpickle(os.path.join(config.workdir, "cat.pkl"))

print momentsml.tools.table.info(cat)


fig = plt.figure(figsize=(20, 13))


#tru_rad = Feature("tru_rad", nicename="True half-light radius [pixel]")
#tru_mag = Feature("tru_mag", 20, 25, nicename="Magnitude")
#tru_sersicn = Feature("tru_sersicn", 0, 7, nicename="True Sersic index")

tru_rad = Feature("tru_rad")
tru_mag = Feature("tru_mag")
tru_sersicn = Feature("tru_sersicn")


ax = fig.add_subplot(2, 3, 1)
momentsml.plot.scatter.scatter(ax, cat, tru_sersicn, tru_rad, sidehists=True, sidehistkwargs={"bins":20})
ax = fig.add_subplot(2, 3, 2)
momentsml.plot.scatter.scatter(ax, cat, tru_mag,  tru_rad, sidehists=True, sidehistkwargs={"bins":20})

ax = fig.add_subplot(2, 3, 3)
s = momentsml.tools.table.Selector("select", [
		("in", "tru_mag", 24.25, 24.75),
	])
bincat = s.select(cat)
momentsml.plot.hist.hist(ax, bincat, tru_rad)
s = momentsml.tools.table.Selector("select", [
		("in", "tru_mag", 22.75, 23.25),
	])
bincat = s.select(cat)
momentsml.plot.hist.hist(ax, bincat, tru_rad, color="blue")



plt.tight_layout()
plt.show()
plt.close(fig) # Helps releasing memory when calling in large loops.





import numpy as np
import math


profile_types = ["Gaussian", "Sersic", "EBulgeDisk"] # Order defines the numerical codes of the galaxy profile types
# EBulgeDisk is a simple Bulge + Disk model that we use for some EC tests.


class Params:
	"""
	A container for the distributions describing the parameters of a simulated galaxy.
	To use it, write your own child class (which inherits from this one) and redefine the
	"draw" method as you wish.
	The distributions provided here are really just examples!
	Of course, you are free to add "correlations" between parameters.
	"""
	
	def __init__(self, name=None):
		"""
		:param name: a simple name for your Params object ("test1", "nonoise", "gdisk", ...).
			This name might be used in filenames by pipelines, so keep it short and whitespace-free.
			If None, I will use the name of the class. This is usually exactly what you want, when
			you make your own class LowNoise()
			
		:type name: string
		
		"""
		if name == None:
			self.name = self.__class__.__name__
		else:	
			self.name = name
				
	
	def __str__(self):
		"""
		The string representation is in fact the name:
		"""
		return "%s" % (self.name)

	
	def stat(self):
		"""
		This method returns the "stationnary" parameters, it gets called only *once* per catalog creation.
		
		If you want to override any of the parameters returned by draw() with some stationnary values, just include
		them in the output of this stat()!
		"""
		
		tru_s1 = 0 # Leaving these lensing parameters to the integers 0 0 1 means that the "lens" method will not get called.
		tru_s2 = 0
		tru_mu = 1
		
		snc_type = 0 # 0 means no shape noise cancellation
		
		return {
			"tru_s1" : tru_s1, # shear component 1, in "g" convention
			"tru_s2" : tru_s2, # component 2
			"tru_mu" : tru_mu, # magnification
			"snc_type" : snc_type, # The type of shape noise cancellation. 0 means none, n means n-fold
		}
		
		
	
	def draw(self, ix, iy, nx, ny):
		"""
		The method that gets called to draw the random variables, and which you shoud redefine.
		In the given example, we only draw Sersic profiles, and we do include parameters for Gaussian PSFs.
		
		:param ix: x index of the galaxy, going from 0 to nx-1
		:param iy: idem for y
		:param nx: number of x indexes
		:param ny: idem for y
	
		As you see, this method will know about the "position" of a galaxy on a grid.
		This allows you to generate non-random distributions, using these grid indexes.
		For instance, you should not draw sersic indices randomly,
		as changing it from stamp to stamp significantly slows down galsim !
		
		.. note:: Comments on what to return are included in the example code of this method.
		
		"""
		
		# See the return statement below for comments!
		
		tru_type = 1
		tru_flux =  np.random.uniform(1000.0, 10000.0)
		
		#tru_sigma = np.random.uniform(1.0, 4.0) # For sersic, we do not have to provide this.
		tru_rad = np.random.uniform(1.0, 4.0)
		
		max_g = 0.6
		g = np.random.triangular(0.0, max_g, max_g) # This triangular g gives uniform disk (if you want that)
		theta = 2.0 * np.pi * np.random.uniform(0.0, 1.0)		
		(tru_g1, tru_g2) = (g * np.cos(2.0 * theta), g * np.sin(2.0 * theta))
			
		tru_sersicn =  0.5 + (float(iy)/float(ny))**2.0 * 3.0
		
		tru_sky_level = 100.0
		tru_gain = 1.0
		tru_read_noise = 3.0
		
		tru_pixel = -1.0
		
		tru_psf_sigma = 1.0
		
		max_psf_g = 0.2
		psf_g = np.random.triangular(0.0, max_psf_g, max_psf_g)
		psf_theta = 2.0 * np.pi * np.random.uniform(0.0, 1.0)		
		(tru_psf_g1, tru_psf_g2) = (psf_g * np.cos(2.0 * psf_theta), psf_g * np.sin(2.0 * psf_theta))
			
		return {
			"tru_type" : tru_type, # Galaxy profile type. 0 is Gaussian, 1 is Sersic (see sim.params.profile_types)
			"tru_flux" : tru_flux, # in ADU
			#"tru_sigma" : tru_sigma, # Size sigma, used for Gaussians
			"tru_rad" : tru_rad, # Half-light radius, used for Sersics
			"tru_sersicn" : tru_sersicn, # Lower sersic index = less concentrated = lower adamom_rho4
			"tru_g1" : tru_g1,
			"tru_g2" : tru_g2,
			
			"tru_sky_level" : tru_sky_level, # in ADU, just for generating noise, will not remain in the image
			"tru_gain" : tru_gain, # in photons/ADU. Make this negative to have no Poisson noise
			"tru_read_noise" : tru_read_noise, # in photons if gain > 0.0, otherwise in ADU.Set this to zero to have no flat Gaussian noise
			
			"tru_pixel" : tru_pixel, # If positive, adds an extra convolution by that many pixels to the simulation process
			
			"tru_psf_sigma" : tru_psf_sigma, # Gaussian PSFs will only be used if no PSF stamps are given
			"tru_psf_g1" : tru_psf_g1, # You don't have to provide those parameters if you will use stamps
			"tru_psf_g2" : tru_psf_g2
		}




"""
Functions to draw grids of simulated galaxies into FITS images.
"""

import numpy as np
import math
import random
import os
import sys
import copy

import logging
logger = logging.getLogger(__name__)

import astropy.table
import galsim
from datetime import datetime

from .. import tools
from . import params


def drawcat(simparams, n=10, nc=2, stampsize=64, pixelscale=1.0, idprefix="", metadict=None):
	"""
	Generates a catalog of all the "truth" input parameters for each simulated galaxy.
	
	:param simparams: a sim.Params instance that defines the distributions of parameters
	:param n: number of galaxies (must be a multiple of nc)
		
	:type n: int
	:param nc: number of "columns" on which I should place these galaxies
		Note that the effective number of "columns" (called nx in the code) might get inflated if you use shape noise cancelation!
	:type nc: int
	:param stampsize: width = height of desired stamps, in pixels
	:type stampsize: int
	:param pixelscale: scale in arcsec of the image, IGNORED: we always draw images with a scale of 1.0
	:type pixelscale: float
	:param idprefix: a string to use as prefix for the galaxy ids. Was tempted to call this idefix.
	
	:param medadict: further content added to the meta of the output catalog
	
	The ix index moves faster than the iy index when iterating over the output catalog.
	So if you want a parameter to change seldomly, link it to the iy index.
	Different SNC-versions of a galaxy are placed directly one after the other, with consecutive ix indices but same iy.
	(however for the purpose of drawing the parameters, the SNC-versions *all* have the same ix as well, of course)
	
	:returns: A catalog (astropy table). The stampsize is stored in meta.
	
	"""
	if int(stampsize)%2 != 0:
		raise RuntimeError("stampsize should be even!")

	if n % nc != 0 or n < nc:
		raise RuntimeError("n must be a multiple of nc!")

	logger.info("Drawing a catalog of %i truly different galaxies using params '%s'..." % (n, simparams))
	statparams = simparams.stat() # We call this only *once*
	logger.info("The params '{simparams.name}' have stat '{0}'".format(statparams, simparams=simparams))
	
	ny = int(int(n)/int(nc)) # number of lines does not change with SNC.
	if statparams["snc_type"] == 0:	
		nsnc = 1 # the number of SNC versions. "No SNC" means the same thing as "1 SNC"...
	elif statparams["snc_type"] > 0:
		nsnc = statparams["snc_type"] # If snc_type is positive, it simply corresponds to the number of SNC rotations!
	else:
		raise RuntimeError("Did not understand snc_type")
	nx = nsnc * nc
	sncrot = 180.0/float(nsnc) # Rotation for SNC, in degrees. For example, 90.0 for snc_type == 2
	
	logger.info("The grid will be %i x %i, and the number of SNC rotations is %i." % (nx, ny, nsnc))
	
	rows = [] # The "table"
	for i in range(n): # We loop over all "truely different" galaxies (not all SNC galaxies)
		
		# The indices used to draw parameters for each of these truely different galaxies:
		(piy, pix) = divmod(i, nc)		
		assert pix < nc and piy < ny
	
		# And draw one:
		gal = simparams.draw(pix, piy, nc, ny)
		
		# Now things get different depending on SNC
		if statparams["snc_type"] == 0:	# No SNC, so we simply add this galaxy to the list.
		
			gal["ix"] = pix
			gal["iy"] = piy
			
			gal.update(statparams) # This would overwrite any of the "draw" params.
			rows.append(gal) # So rows will be a list of dicts
		
		
		else: # SNC with nsnc different versions rotated by sncrot degrees
		
			for roti in range(nsnc):
				rotgal = copy.deepcopy(gal)
				
				# We perform the rotation in different ways, depending on the parametrisation
				profile_type = params.profile_types[rotgal["tru_type"]]
				if profile_type in ["Sersic", "Gaussian"]:
					(rotgal["tru_g1"], rotgal["tru_g2"]) = tools.calc.rotg(gal["tru_g1"], gal["tru_g2"], roti*sncrot)
				elif profile_type is "EBulgeDisk":
					rotgal["tru_theta"] = gal["tru_theta"] + roti*sncrot
				else:
					raise RuntimeError("Unknown profile type")
				
				# And now each of the SNC versions gets a consecutive ix but the same iy:
				rotgal["ix"] = pix * nsnc + roti
				rotgal["iy"] = piy
				
				rotgal.update(statparams)
				rows.append(rotgal)
				#print rotgal
		
		
	# A second loop simply adds the pixel positions and ids, for all galaxies (not just truely different ones):
	for (i, gal) in enumerate(rows):
		gal["id"] = idprefix + str(i)
		gal["x"] = gal["ix"]*stampsize + stampsize/2.0 + 0.5 # I'm not calling this tru_x, as it will be jittered, and also as a simple x is default.
		gal["y"] = gal["iy"]*stampsize + stampsize/2.0 + 0.5
			
				
	# There are many ways to build a new astropy.table
	# One of them directly uses a list of dicts...
	
	catalog = astropy.table.Table(rows=rows)
	logger.info("Drawing of catalog done")
	
	# The following is aimed at drawimg:
	catalog.meta["stampsize"] = stampsize
	catalog.meta["pixelscale"] = pixelscale
	
	# Checking the catalog length
	assert len(catalog) == nsnc * n
	
	# Adding some useful stuff to the meta:
	catalog.meta["nx"] = nx
	catalog.meta["ny"] = ny
	catalog.meta["nsnc"] = nsnc
	catalog.meta["snc_type"] = statparams["snc_type"]
	
	if metadict:
		catalog.meta.update(metadict)
	
	return catalog
	



def drawimg(catalog, simgalimgfilepath="test.fits", simtrugalimgfilepath=None, simpsfimgfilepath=None, gsparams=None, sersiccut=None):

	"""
	Turns a catalog as obtained from drawcat into FITS images.
	Only the position jitter and the pixel noise are randomized. All the other info is taken from the input catalog.
	So simply call me several times for the same input to get different realizations of the same galaxies.
	To specify the PSFs, add a meta["psf"] ImageInfo object to the catalog.
		
	:param catalog: an input catalog of galaxy shape parameters, as returned by drawcat.
		The corresponding stampsize must be provided as catalog.meta["stampsize"].
		If you specify a psf image in catalog.meta["psf"], your catalog must of course also contain
		PSF coordinates for that image.
		If no PSF stamps are specifed, the code will look for Gaussian PSF parameters in the catalog.
		If such parameters are not given, no PSF convolution is done.
		
	:param simgalimgfilepath: where I write my output image of simulated and noisy galaxies
	:param simtrugalimgfilepath: (optional) where I write the image without convolution and noise
	:param simpsfimgfilepath: (optional) where I write the PSFs
	
	:param sersiccut: cuts the sersic profile at this number of rad
	
	.. note::
		See this function in MomentsML v4 (great3) for attempts to speed up galsim by playing with fft params, accuracy, etc...
	
	.. note::
		About speed, if you specify trunc, better express the scale radius.
		But scale radius is crazy dependent on n, so I keep half-light-radius
		http://galsim-developers.github.io/GalSim/classgalsim_1_1base_1_1_sersic.html
		
		
	.. note::
		To use the hacks, give "metadict":{"hack":"nicobackgals"} as drawcatkwargs to sim.run.multi()...
	
	"""
	starttime = datetime.now()	
	
	hack = catalog.meta.get("hack", None)
	
	if hack is None: # The default behavior, without specific gsparams or tricks.
	
		if gsparams is None:
			gsparams = galsim.GSParams(maximum_fft_size=10240)

		if "nx" not in catalog.meta.keys() or "ny" not in catalog.meta.keys():
			raise RuntimeError("Provide nx and ny in the meta data of the input catalog to drawimg.")
		if "stampsize" not in catalog.meta.keys():
			raise RuntimeError("Provide stampsize in the meta data of the input catalog to drawimg.")	
		
		nx = catalog.meta["nx"]
		ny = catalog.meta["ny"]
		stampsize = catalog.meta["stampsize"] # The stamps I'm going to draw
			
		logger.info("Drawing images of %i galaxies on a %i x %i grid..." % (len(catalog), nx, ny))
		logger.info("The stampsize for the simulated galaxies is %i." % (stampsize))
	
		# A special function checks the combination of settings in the provided catalog:
		todo = checkcat(catalog)
		
		if "loadpsfimg" in todo:
			psfimg = catalog.meta["psf"].load() # Loading the actual GalSim Image
			psfinfo = catalog.meta["psf"]
	
		if "tru_pixel" in todo:
			# This is only if you want "effective pixels" larger than the actual pixels (related to SBE, normally you don't want this).
			pix = galsim.Pixel(catalog["tru_pixel"][0]) # We have checked in checkcat that all values are equal.
			
		# Galsim random number generators
		rng = galsim.BaseDeviate()
		ud = galsim.UniformDeviate() # This gives a random float in [0, 1)

		# We prepare the big images :
		gal_image = galsim.ImageF(stampsize * nx , stampsize * ny)
		trugal_image = galsim.ImageF(stampsize * nx , stampsize * ny)
		psf_image = galsim.ImageF(stampsize * nx , stampsize * ny)
		
		gal_image.scale = 1.0 # we use pixels as units. Note that if you change something here, you also have to change the jitter.
		trugal_image.scale = 1.0
		psf_image.scale = 1.0

		# And loop through the catalog:
		for row in catalog:
			
			# Some simplistic progress indication:
			fracdone = float(row.index) / len(catalog)
			if row.index%500 == 0:
				logger.info("%6.2f%% done (%i/%i) " % (fracdone*100.0, row.index, len(catalog)))
			
			# We will draw this galaxy in a postage stamp, but first we need the bounds of this stamp.
			ix = int(row["ix"])
			iy = int(row["iy"])
			assert ix < nx and iy < ny
			bounds = galsim.BoundsI(ix*stampsize+1 , (ix+1)*stampsize, iy*stampsize+1 , (iy+1)*stampsize) # Default Galsim convention, index starts at 1.
			gal_stamp = gal_image[bounds]
			trugal_stamp = trugal_image[bounds]
			psf_stamp = psf_image[bounds]
	
			# We draw the desired profile
			profile_type = params.profile_types[row["tru_type"]]
			
			if profile_type == "Sersic":
				if sersiccut is None:
					trunc = 0
				else:
					trunc = float(row["tru_rad"]) * sersiccut
				gal = galsim.Sersic(n=float(row["tru_sersicn"]), half_light_radius=float(row["tru_rad"]), flux=float(row["tru_flux"]), gsparams=gsparams, trunc=trunc)
				# We make this profile elliptical
				gal = gal.shear(g1=row["tru_g1"], g2=row["tru_g2"]) # This adds the ellipticity to the galaxy

			elif profile_type == "Gaussian":
				
				gal = galsim.Gaussian(flux=float(row["tru_flux"]), sigma=float(row["tru_sigma"]), gsparams=gsparams)
				# We make this profile elliptical
				gal = gal.shear(g1=row["tru_g1"], g2=row["tru_g2"]) # This adds the ellipticity to the galaxy	
	
			elif profile_type == "EBulgeDisk":
			
				# A more advanced Bulge + Disk model
				# It needs GalSim version master, as of April 2017 (probably 1.5).
				
				# Get a Sersic bulge:
				bulge = galsim.Sersic(n=row["tru_bulge_sersicn"], half_light_radius=row["tru_bulge_hlr"], flux=row["tru_bulge_flux"])
				# Make it elliptical:
				bulge_ell = galsim.Shear(g=row["tru_bulge_g"], beta=row["tru_theta"] * galsim.degrees)
				bulge = bulge.shear(bulge_ell)
				
				# Get a disk
				scale_radius = row["tru_disk_hlr"] / galsim.Exponential._hlr_factor
				disk = galsim.InclinedExponential(inclination=row["tru_disk_tilt"] * galsim.degrees, scale_radius=scale_radius,
					flux=row["tru_disk_flux"], scale_h_over_r=row["tru_disk_scale_h_over_r"])
				# Rotate it in the same orientation as the bulge:
				disk = disk.rotate(row["tru_theta"] * galsim.degrees)
				
				# And we add those profiles, as done in GalSim demo3.py :
				gal = bulge + disk
				
			else:
				raise RuntimeError("Unknown galaxy profile!")	
		
		
			# And now we add lensing, if s1, s2 and mu are different from no lensing...
			if row["tru_s1"] != 0 or row["tru_s2"] != 0 or row["tru_mu"] != 1:
				gal = gal.lens(float(row["tru_s1"]), float(row["tru_s2"]), float(row["tru_mu"]))
			else:
				pass
				#logger.info("No lensing!")
			
			
			# We apply some jitter to the position of this galaxy
			xjitter = ud() - 0.5 # This is the minimum amount -- should we do more, as real galaxies are not that well centered in their stamps ?
			yjitter = ud() - 0.5
			gal = gal.shift(xjitter,yjitter)
			
			# We draw the pure unconvolved galaxy
			if simtrugalimgfilepath != None:
				gal.drawImage(trugal_stamp, method="auto") # Will convolve by the sampling pixel.	

			# We prepare/get the PSF and do the convolution:
			
			# Should the final operation skip the convolution by the pixel (because the PSF already is in large pixels) ?
			skip_pixel_conv = False
			
			if "usegausspsf" in todo:
				
				if row["tru_psf_sigma"] < 0.0:
					raise RuntimeError("Unknown hack!")	
				else:
					psf = galsim.Gaussian(flux=1., sigma=row["tru_psf_sigma"])	
					psf = psf.shear(g1=row["tru_psf_g1"], g2=row["tru_psf_g2"])
			
					# Let's apply some jitter to the position of the PSF (not sure if this is required, but should not harm ?)
					psf_xjitter = ud() - 0.5
					psf_yjitter = ud() - 0.5
					psf = psf.shift(psf_xjitter,psf_yjitter)
					
				if simpsfimgfilepath != None:
					psf.drawImage(psf_stamp, method="auto") # Will convolve by the sampling pixel.
		
				if "tru_pixel" in todo: # Not sure if this should only apply to gaussian PSFs, but so far this seems OK.
					# Remember that this is an "additional" pixel convolution, not the usual sampling-related convolution that happens in drawImage.
					galconv = galsim.Convolve([gal, psf, pix])
				
				else:
					galconv = galsim.Convolve([gal,psf])
				
			elif "loadpsfimg" in todo:
				
				(inputpsfstamp, flag) = tools.image.getstamp(row[psfinfo.xname], row[psfinfo.yname], psfimg, psfinfo.stampsize)
				psfpixelscale = getattr(psfinfo, "pixelscale", 1.0) # Using getattr so that it works with old objects as well
				if psfpixelscale > 0.5:
					#logger.warning("You seem to be using a sampled PSF with large pixels (e.g., observed stars). I'll do my best and skip the pixel conv, but this might well lead to errors.")
					skip_pixel_conv = True
				if flag != 0:
					raise RuntimeError("Could not extract a %ix%i stamp at (%.2f, %.2f) from the psfimg %s" %\
						(psfinfo.stampsize, psfinfo.stampsize, row[psfinfo.xname], row[psfinfo.yname], psfinfo.name))
				psf = galsim.InterpolatedImage(inputpsfstamp, flux=1.0, scale=psfpixelscale)
				if simpsfimgfilepath != None:
					psf.drawImage(psf_stamp, method="no_pixel") # psf_stamp has a different size than inputpsfstamp, so this could lead to problems one day.
				
				#galconv = galsim.Convolve([gal,psf], real_space=False)	
				galconv = galsim.Convolve([gal,psf])			

			elif "nopsf" in todo:
				# Nothing to do		
				galconv = gal
				
			else:
				raise RuntimeError("Bug in todo.")
		
			# Draw the convolved galaxy	
			if skip_pixel_conv == False:	
				galconv.drawImage(gal_stamp, method="auto") # This will convolve by the image sampling pixel. Don't do this yourself ahead! 
			else:
				#logger.warning("NOT computing any pixel convolution")
				galconv.drawImage(gal_stamp, method="no_pixel") # Simply uses pixel-center values. Know what you are doing, see doc of galsim. 

		
			# And add noise to the convolved galaxy:
			gal_stamp.addNoise(galsim.CCDNoise(rng, sky_level=float(row["tru_sky_level"]), gain=float(row["tru_gain"]), read_noise=float(row["tru_read_noise"])))
	
		
		logger.info("Done with drawing, now writing output FITS files ...")	

		gal_image.write(simgalimgfilepath)
		
		if simtrugalimgfilepath != None:
			trugal_image.write(simtrugalimgfilepath)
		
		if simpsfimgfilepath != None:
			psf_image.write(simpsfimgfilepath)
		
			
	elif hack == "nicobackgals":
		"""
		This is taken and/or adapted from Nico's simulation code, to be sure to do the same thing.
			- Nico uses GalSim in units of arcsec (instaed of pixels), and we suspect that this affects the convolution.
			- Nico has different gsparams
			- Nico truncates Sersic profiles
			
		"""
		logger.info("Using special hack for nicobackgals")
						
		gsparams = galsim.GSParams(xvalue_accuracy=2.e-4, kvalue_accuracy=2.e-4, maxk_threshold=5.e-3, folding_threshold=1.e-2)
		pixel_scale = 0.1
		
		if "nx" not in catalog.meta.keys() or "ny" not in catalog.meta.keys():
			raise RuntimeError("Provide nx and ny in the meta data of the input catalog to drawimg.")
		if "stampsize" not in catalog.meta.keys():
			raise RuntimeError("Provide stampsize in the meta data of the input catalog to drawimg.")	
		
		nx = catalog.meta["nx"]
		ny = catalog.meta["ny"]
		stampsize = catalog.meta["stampsize"] # The stamps I'm going to draw
			
		logger.info("Drawing images of %i galaxies on a %i x %i grid..." % (len(catalog), nx, ny))
		logger.info("The stampsize for the simulated galaxies is %i." % (stampsize))

		# Galsim random number generators
		rng = galsim.BaseDeviate()
		ud = galsim.UniformDeviate() # This gives a random float in [0, 1)

		# We prepare the big image, and set the pixel scale
		gal_image = galsim.ImageF(stampsize * nx , stampsize * ny, scale=pixel_scale)
		

		psf = galsim.Airy(lam=800, diam=1.2, obscuration=0.3, scale_unit=galsim.arcsec, flux=1./3) + \
			galsim.Airy(lam=700, diam=1.2, obscuration=0.3, scale_unit=galsim.arcsec, flux=1./3) + \
			galsim.Airy(lam=600, diam=1.2, obscuration=0.3, scale_unit=galsim.arcsec, flux=1./3)

		for row in catalog:
			
			# Some simplistic progress indication:
			fracdone = float(row.index) / len(catalog)
			if row.index%500 == 0:
				logger.info("%6.2f%% done (%i/%i) " % (fracdone*100.0, row.index, len(catalog)))
			
			# We will draw this galaxy in a postage stamp, but first we need the bounds of this stamp.
			ix = int(row["ix"])
			iy = int(row["iy"])
			assert ix < nx and iy < ny
			bounds = galsim.BoundsI(ix*stampsize+1 , (ix+1)*stampsize, iy*stampsize+1 , (iy+1)*stampsize) # Default Galsim convention, index starts at 1.
			gal_stamp = gal_image[bounds]
	
	
			# Drawing the galaxy
			half_light_radius=float(row["tru_rad"]) * pixel_scale
			gal = galsim.Sersic(n=float(row["tru_sersicn"]), half_light_radius=half_light_radius, flux=float(row["tru_flux"]), gsparams=gsparams, trunc=half_light_radius*4.5)
			# We make this profile elliptical
			gal = gal.shear(g1=row["tru_g1"], g2=row["tru_g2"]) # This adds the ellipticity to the galaxy

			# And now we add lensing, if s1, s2 and mu are different from no lensing...
			if row["tru_s1"] != 0 or row["tru_s2"] != 0 or row["tru_mu"] != 1:
				gal = gal.lens(float(row["tru_s1"]), float(row["tru_s2"]), float(row["tru_mu"]))
			else:
				pass
				#logger.info("No lensing!")
			
			# We apply some jitter to the position of this galaxy # it seems this has to be given in the image scale...
			xjitter = (ud() - 0.5) * pixel_scale
			yjitter = (ud() - 0.5) * pixel_scale
			gal = gal.shift(xjitter,yjitter)

			final = galsim.Convolve([psf, gal])

 			final.drawImage(gal_stamp)

 			# And add noise to the convolved galaxy:
			gal_stamp.addNoise(galsim.CCDNoise(rng, sky_level=float(row["tru_sky_level"]), gain=float(row["tru_gain"]), read_noise=float(row["tru_read_noise"])))
	
			
		logger.info("Done with drawing, now writing output FITS files ...")	

		gal_image.write(simgalimgfilepath)
		
	
			
	else:
		raise RuntimeError("Unknown hack")

	
	
	endtime = datetime.now()
	logger.info("This drawing took %s" % (str(endtime - starttime)))



def checkcat(cat):
	"""
	Checks the input catalog for consistency, compliance, and purpose (-> "to do"...)
	This function is mainly here to make things fail nicely if something is missing.
	It returns a list containing flags that will control some aspects of drawimg.
	"""
	
	todo = [] # The output list.
	
	# We check some fields that should always be there:
	for f in ["tru_type", "tru_sky_level", "tru_gain", "tru_read_noise"]:
		if f not in cat.colnames:
			raise RuntimeError("The field '%s' is not in the catalog (i.e., the simulation parameters)!" % (f))

	# What profiles do we have in there ?
	contained_profile_types = list(set([params.profile_types[row["tru_type"]] for row in cat]))
	logger.info("Galaxy profile types to be simulated: %s" % (str(contained_profile_types)))	

	# We check the additional required fields for each of these profile types
	if "Gaussian" in contained_profile_types:
		for f in ["tru_sigma", "tru_g1", "tru_g2"]:
			if f not in cat.colnames:
				raise RuntimeError("I should draw a Gaussian profile, but '%s' is not in the catalog (i.e., the simulation parameters)!" % (f))
	if "Sersic" in contained_profile_types:
		for f in ["tru_rad", "tru_sersicn", "tru_g1", "tru_g2"]:
			if f not in cat.colnames:
				raise RuntimeError("I should draw a Sersic profile, but '%s' is not in the catalog (i.e., the simulation parameters)!" % (f))
	if "EBulgeDisk" in contained_profile_types:
		for f in ["tru_theta", "tru_bulge_g", "tru_bulge_sersicn", "tru_bulge_hlr", "tru_bulge_flux", "tru_disk_tilt", "tru_disk_scale_h_over_r", "tru_disk_hlr", "tru_disk_flux"]:
			if f not in cat.colnames:
				raise RuntimeError("I should draw a Sersic profile, but '%s' is not in the catalog (i.e., the simulation parameters)!" % (f))

		
	# And now we check the information provided about the PSFs to be used
	
	if "psf" in cat.meta: # The user provided some PSFs in form of stamps
		logger.info("I will use provided PSF stamps (from '%s')" % (str(cat.meta["psf"])))
		cat.meta["psf"].checkcolumns(cat) # Check that the ImageInfo object matches to the catalog.
		
		todo.append("loadpsfimg")
				
	else: # I will look for PSF params in the catalog, and use analytical Gaussians
		
		psf_gauss_ok = True
		for f in ["tru_psf_sigma", "tru_psf_g1", "tru_psf_g2"]:
			if f not in cat.colnames:
				psf_gauss_ok = False
		
		if psf_gauss_ok:
			logger.info("I will use analytical PSFs.")
			todo.append("usegausspsf")
		else:
			logger.warning("No or not enough PSF information given, I will NOT convolve the simulated galaxies!")
			todo.append("nopsf")
	
	
	# Now let's have a look at the extra pixel convolution (added with SBE in mind)
	
	if "tru_pixel" in cat.colnames:
		tru_pixels = list(set(cat["tru_pixel"]))
		if len(tru_pixels) != 1:
			raise RuntimeError("You're mixing seveal tru_pixel values, probably a mistake.") # If you ever change this, fix code that currently generates only one pix!
		tru_pixel = tru_pixels[0]
		if tru_pixel > 0.0:
			logger.warning("The galaxy profiles will be convolved with an extra {0:.1f} pixels".format(tru_pixel))
		
			todo.append("tru_pixel")
	
	return todo



"""
The ImageInfo class
"""

import sys
import os

import image

import logging
logger = logging.getLogger(__name__)

class ImageInfo():
	"""
	Class to structure meta-data related to an image (stored in a FITS file) inside the meta dict of a catalog (astropy.table) object.
	Putting objects of this class in a catalogs's meta dict "links" an image file to that catalog.
	This way a catalog can have links to many images (galaxy stamps, PSF stamps, multiple exposures, filters...)
	
	When working with simple single-exposure images, the convention that shape measurement functions are asked to understand is that
	the galaxy stamp image is described in catalog.meta["img"], and (optionally) the PSF stamp image is in catalog.meta["psf"].
	We can derive a more sophisticated hierarchical structure when needed.
	"""

	def __init__(self, filepath, xname="x", yname="y", stampsize=None, name=None, workdir=None, pixelscale=1.0):
		"""
		:param filepath: absolute path to the FITS image
		:type filepath: string
		:param xname: catalog column name containing the x coordinate of the center of the stamp
		:type xname: int
		:param yname: idem for y
		:type yname: int
		:param stampsize: width and height of a stamp, in pixels, if the image consists of gridded stamps.
			Leave it to None if the image does not hold stamps (real life images, for instance).
		:type stampsize: int
		:param name: a name for the image. By default (None) the filename will be used, but sometimes
			it might be helpful to specify different names. Functions use these names to refer to the image in logs etc.
		:type name: string
		:param workdir: path to a working directory for this image
		:type workdir: string
		:param pixelscale: For PSF images: size of the pixels, in "arcsec". In fact, arcsec does not matter -- we never use angles.
			Only the relative value with respect to other ImageInfos (that have a scale of 1.0) matters. It allows for instance to use supersampled
			PSF images, with say 5 times "smaller" pixels than the detector pixels (set it to 0.2 in this case).
		:type pixelscale: float
		"""
		
		self.filepath = filepath
		self.xname = xname
		self.yname = yname
		self.stampsize = stampsize
		self.pixelscale = pixelscale

		
		if name is None:
			self.name = os.path.splitext(os.path.basename(filepath))[0]
		else:
			self.name = name
		
		self.workdir = workdir # To be developed...
				

	def __str__(self):
		"""
		Retruns a short description of this object
		"""
		return "ImageInfo(%s, %s, %s, %s, %s)" % (self.name, self.xname, self.yname, self.stampsize, self.pixelscale)

	def __repr__(self):
		"""
		Returns str(self) -- convenient when printing catalog.meta.
		"""
		return self.__str__()

	def load(self):
		"""
		Returns the image as GalSim Image object.
		"""
		return image.loadimg(self.filepath)


	def checkcolumns(self, catalog):
		"""
		Verifies that the attributes xname and yname are available as columns in the given catalog.
		If not, raises a nicely descriptive RuntimeError.
		As this is such a common check, a dedicated method seems appropriate.
		"""
		if not (self.xname in catalog.colnames and self.yname in catalog.colnames):
			raise RuntimeError("The columns (%s, %s) are not among the ones available in the catalog: %s"\
				% (self.xname, self.yname, catalog.colnames))

		
	def get_stampsize(self, stampsize=None):
		"""
		Returns the stampsize to be used for a shape measurement.
		If the stampsize is specified as argument, this supersedes the object's stampsize.
		If both the object's stampsize and the argument are different from None,
		warns the user if they don't agree.
		If both are None, raises an exception.
		
		:param stampsize: a stampsize given e.g. as argument to a measfct 
		 
		"""
		
		# First guess, the own stampsize:
		use_stampsize = self.stampsize
		
		# If given, the argument supersedes the first guess:
		if stampsize != None:
			use_stampsize = stampsize
			if self.stampsize != None: # means that both were given:
				if stampsize != self.stampsize:
					logger.warning("A stampsize %i was specified, but stamps seem to have a stampsize of %i" %\
						(stampsize, self.stampsize))
		
		# If no stampsize has been found:
		if use_stampsize is None:
			raise RuntimeError("Specify a stampsize, either in the ImageInfo object or as argument")
		
		return use_stampsize
		
		
		



"""
Similar to test_mult_learn, but this time as inverse regression.

Training z = x * y

x and y are noisy inputs.
"""

import numpy as np
import tenbilac
import os

import logging
logging.basicConfig(level=logging.INFO)


nrea = 5
ncas = 500

# We prepare the data:

# Range of x and y
xs = np.random.uniform(-3, 6, ncas)
ys = np.random.uniform(-10, -5, ncas)

zs = xs * ys 

inputs = np.array([xs, ys])
inputs = np.tile(inputs, (nrea, 1, 1))
# We add noise:
inputs += 0.001*np.random.randn(inputs.size).reshape(inputs.shape)

# This is 3D (rea, features=2, case)
#print inputs
print inputs.shape

targets = np.array([zs])

# This is 2D (feature=1, case)
#print targets
print targets.shape


# Now the normer:

inputnormer = tenbilac.data.Normer(inputs, type="sa1")
inputs = inputnormer(inputs)
targetnormer = tenbilac.data.Normer(targets, type="sa1")
targets = targetnormer(targets)

print inputnormer
print targetnormer


dat = tenbilac.data.Traindata(inputs=inputs, targets=targets)

net = tenbilac.net.Net(ni=2, nhs=[-1], no=1, actfctname="iden", oactfctname="iden",  multactfctname="iden", inames=["x", "y"], onames=["z"])

# The exact solution:
net.setidentity()
#net.layers[0].weights[0,1] = 1.0
#net.layers[1].weights[0] = ((inputnormer.b[0]*inputnormer.b[1])/targetnormer.b[0])

net.addnoise(multwscale=0.5, wscale=0.1, bscale=0.1)


print net.report()


training = tenbilac.train.Training(net, dat, errfctname="msb", autoplot=True, trackbiases=True, autoplotdirpath=".")


training.opt(algo="brute", mbsize=None, mbfrac=1.0, mbloops=1, maxiter=15, gtol=1e-8)
training.opt(algo="bfgs", mbsize=None, mbfrac=1.0, mbloops=1, maxiter=15, gtol=1e-8)

print net.report()


outs = targetnormer.denorm(net.predict(inputs))
#outs = net.predict(inputs)
# Shape is (rea, neuron, case)

assert zs.size == ncas
residues = outs[:,0,:] - zs # We have only one neuron
# Shape is (rea, case)

# Stats for each case, over the realizations:
meanres = np.mean(residues, axis=0)
stdres = np.std(residues, axis=0)
assert meanres.size == ncas


#training.save("test.pkl", keepdata=True)
#exit()
import matplotlib.pyplot as plt


fig = plt.figure(figsize=(12, 5))

plt.subplot(1, 2, 1)
plt.scatter(xs, ys, c=meanres, s=80, marker="o", edgecolors="face")
plt.xlabel("x")
plt.ylabel("y")
plt.colorbar()
plt.title("Mean residue")

plt.subplot(1, 2, 2)
plt.scatter(xs, ys, c=stdres, s=80, marker="o", edgecolors="face")
plt.xlabel("x")
plt.ylabel("y")
plt.colorbar()
plt.title("Std residue")

plt.tight_layout()


plt.show()




import matplotlib
matplotlib.use("AGG")


import momentsml.tools
import momentsml.learn
import momentsml
import tenbilac

import config
import numpy as np
import os

from momentsml.tools.feature import Feature
import matplotlib.pyplot as plt

import plot_2_val

import logging
logging.basicConfig(format=config.loggerformat, level=logging.DEBUG)
logger = logging.getLogger(__name__)


select = True


for subfield in config.great3.subfields:
	
	
	logger.info("Working on subfield {}".format(subfield))
	
	# Getting the path to the correct directories
	measdir = config.great3.subpath(subfield, "simmeas", config.datasets["train-shear"])
	traindir = config.great3.subpath(subfield, "ml", config.datasets["train-shear"])
	
	# And to the catalogue
	traincatpath = os.path.join(measdir, "groupmeascat.pkl")
	cat = momentsml.tools.io.readpickle(traincatpath)
	#print momentsml.tools.table.info(cat)
	
	if select:
		s = momentsml.tools.table.Selector("fortrain", [
			("max", "adamom_failfrac", 0.1),
		])
		cat = s.select(cat)
	
	#exit()	
	#print momentsml.tools.table.info(cat)
	
	
	# Running the training
	dirnames = momentsml.learn.tenbilacrun.train(cat, config.shearconflist, traindir)
	
	# And saving a summary plot next to the tenbilac working directories
	for dirname in dirnames:
		tenbilacdirpath = os.path.join(traindir, dirname)
		ten = tenbilac.com.Tenbilac(tenbilacdirpath)
		ten._readmembers()
		tenbilac.plot.summaryerrevo(ten.committee, filepath=os.path.join(traindir, "{}_summary.png".format(dirname)))
	
	
	#dirnames = ["ada4s1_sum55"]
	#dirnames = momentsml.learn.tenbilacrun.confnames(conflist)
	# Self-predicting
	
	cat = momentsml.tools.io.readpickle(traincatpath) # To get the full catalog with all cases, not just the selected ones
	for (dirname, conf) in zip(dirnames, config.shearconflist):

		predcat = momentsml.learn.tenbilacrun.predict(cat, [conf], traindir)
		predcatpath = os.path.join(traindir, "{}_selfpred.pkl".format(dirname))
		momentsml.tools.io.writepickle(predcat, predcatpath)
		figpredcatpath = os.path.join(traindir, "{}_selfpred.png".format(dirname))
		
		if "s2" in conf[0] or "g2" in conf[0]:
			component = 2
		else:
			component = 1
		plot_2_val.plot(predcat, component, mode=config.trainmode, filepath=figpredcatpath)
	
	
	



"""
Activation functions
"""

import numpy as np

import logging
logger = logging.getLogger(__name__)

def sig(x):
	return 1.0 / (1.0 + np.exp(-x))	# The actual sigmoid
	
def sigi(x):
	return 1.0 / (1.0 + np.exp(-x*4.0))	# A sigmoid with a central slope of 1
			
def sige(x):
	return 2.0 / (1.0 + np.exp(-x)) - 1.0 # Making it even
		
def tanh(x):
	return np.tanh(x)
	
def iden(x):
	return x

def relu(x):
	return np.amax([x, np.zeros_like(x)], axis=0)

def relu6(x):
	return np.amin([np.amax([x, np.zeros_like(x)], axis=0), 6 * np.ones_like(x)], axis=0)


if __name__ == "__main__":
	
	import matplotlib.pyplot as plt
	
	x = np.linspace(-5, 5, 1000)
	acts = [sig, sigi, sige, tanh, iden, relu]
	
	for act in acts:
		plt.plot(x, act(x), label=act.__name__)
		
	plt.xlabel(r"$x$")
	plt.ylabel(r"$f(x)$")
	plt.title("Activation functions")
	plt.ylim(-1.2, 1.2)
	plt.legend()
	plt.grid()
	plt.show()
	


#class Sig:
#	"""
#	Sigmoid	
#	"""
#
#	def __call__(self, x):
#		#return 1.0 / (1.0 + np.exp(-x))	# The actual sigmoid
#		return 2.0 / (1.0 + np.exp(-x)) - 1.0	# Making it even
#		
#		
#
#class Tanh:
#	"""
#	Tanh	
#	"""
#
#	def __call__(self, x):
#		return np.tanh(x)
#	
#	
#class Id:
#	"""
#	Plain identity
#	"""
#	
#	def __call__(self, x):
#		return x
#
#
#if __name__ == "__main__":
#	
#	import matplotlib.pyplot as plt
#	
#	x = np.linspace(-5, 5, 100)
#	acts = [Sig, Tanh, Id]
#	
#	for act in acts:
#		f = act()
#		plt.plot(x, f(x), label=act.__name__)
#		
#	plt.xlabel(r"$x$")
#	plt.ylabel(r"$f(x)$")
#	plt.title("Activation functions")
#	plt.ylim(-1.2, 1.2)
#	plt.legend()
#	plt.grid()
#	plt.show()
	
	



import momentsml.sim
import numpy as np
import random # np.random.choice is only available for newer numpys...
import scipy.stats
import config

import itertools
import galsim

import measfcts
import os

import logging
logging.basicConfig(format=config.loggerformat, level=logging.DEBUG)
logger = logging.getLogger(__name__)

def trunc_rayleigh(sigma, max_val):
	"""
	A truncated Rayleigh distribution
	"""
	assert max_val > sigma
	tmp = max_val + 1.0
	while tmp > max_val:
		tmp = np.random.rayleigh(sigma)
	return tmp

def trunc_gaussian(m, s, minval, maxval):
	"""
	A truncated Gaussian distrib
	"""
	assert maxval > minval
	a, b = (float(minval) - float(m)) / float(s), (float(maxval) - float(m)) / float(s)
	distr = scipy.stats.truncnorm(a, b, loc=m, scale=s)
	return distr.rvs()



sourcecat = momentsml.tools.io.readpickle(config.sourcecat)

	
pixel_scale = 0.1 # arcsec per pixel
exptime = 3.0*565.0	# seconds

gain = 3.1 # electrons/ADU
ron = 4.2 # electrons

skyback = 22.35 # mag per arcsec2, dominated by zodiacal light
zeropoint = 24.6 # mag, VIS instrumental


class EuclidLike(momentsml.sim.params.Params):

	
	def __init__(self, name=None, snc_type=1, shear=0, noise_level=1.0, dist_type="gems"):
		"""
		- snc_type is the number of shape noise cancellation rotations
		- shear is the maximum shear to be drawn, 0 for no shear
		- noise_level 1.0 means that the fiducial noise level will be used, 0 means no noise
		"""
		
		momentsml.sim.params.Params.__init__(self)
		if name is not None:
			self.name = name
		self.snc_type = snc_type
		self.shear = shear
		self.noise_level = noise_level
		self.dist_type = dist_type
		


	def draw_constants(self):
		"""
		Settings that don't change from stamp to stamp
		"""
		
		tru_type = 1 # Sersic
		snc_type = self.snc_type # The type of shape noise cancellation. 0 means none, n means n-fold
		
		tru_sky_level = (pixel_scale)**2 * (exptime/gain) * 10**(-0.4*(skyback - zeropoint)) # in ADU, just for generating noise, will not remain in the image
		
		tru_gain = gain
		tru_read_noise = ron # in e-, given that gain is > 0
		
		tru_pixel = -1.0 # If positive, adds an extra convolution by that many pixels to the simulation process
		
		return {"snc_type":snc_type, "tru_type":tru_type, "tru_sky_level":tru_sky_level,
			"tru_gain":tru_gain, "tru_read_noise":tru_read_noise, "tru_pixel":tru_pixel,
			}


	def draw_s(self):
		"""
		Shear and magnification
		"""
		
		if self.shear > 0.0 and self.shear < 10.0:
			tru_s1 = np.random.uniform(-self.shear, self.shear)
			tru_s2 = np.random.uniform(-self.shear, self.shear)		
		else:
			tru_s1 = 0.0
			tru_s2 = 0.0
		
		tru_mu = 1.0
		return {"tru_s1":tru_s1, "tru_s2":tru_s2, "tru_mu":tru_mu}


	def draw_psf(self):
		"""
		Draws the parameters of the PSF : not used, as , we use an image psf.
		"""
		
		tru_psf_sigma = -1.0
		tru_psf_g1 = 0.0
		tru_psf_g2 = 0.0
		
		return {"tru_psf_sigma":tru_psf_sigma, "tru_psf_g1":tru_psf_g1, "tru_psf_g2":tru_psf_g2}


	def stat(self):
		"""
		Called for each catalog (stat is for stationnary)
		"""
		return self.draw_constants()



	def draw(self, ix, iy, nx, ny):
		"""
		Called for each galaxy	
		"""
		
		#tru_e = trunc_rayleigh(0.25, 0.6)
		#tru_e2 = clip_gaussian(sigma, maxamp)
		#shear = galsim.Shear(e1=tru_e1, e2=tru_e2)
		#tru_g1 = shear.g1
		#tru_g2 = shear.g2
		
		tru_g = trunc_rayleigh(0.25, 0.7) # Follows Hoekstra et al. 2017 and 2015, for the sigma-e (yes, the do use a-b/a+b as def)
		tru_theta = 2.0 * np.pi * np.random.uniform(0.0, 1.0)
		(tru_g1, tru_g2) = (tru_g * np.cos(2.0 * tru_theta), tru_g * np.sin(2.0 * tru_theta))
		
		tru_mag_a = 25.0 - iy * 0.5 - 0.05
		tru_mag_b = 25.0 - iy * 0.5 + 0.05
		found_gal = False
		while found_gal == False:
			source_row = random.choice(sourcecat)
			if source_row["tru_mag"] > tru_mag_a and source_row["tru_mag"] < tru_mag_b :
				found_gal = True
		
		tru_rad = source_row["tru_rad"]
		tru_mag = source_row["tru_mag"]
		tru_sersicn_tmp = source_row["tru_sersicn"]
			
		
		tru_flux =  (exptime / gain) * 10**(-0.4*(tru_mag - zeropoint))
		
		# We want a discrete number of sersic indices to keep GalSim efficient, so we approximate it to the closest step:
		tru_sersicns = np.linspace(0.3, 6.0, 21)
		tru_sersicn = tru_sersicns[(np.abs(tru_sersicns-tru_sersicn_tmp)).argmin()]
		
		out = { # It's ugly to not directly fill this dict, but this makes it clearer what is actually returned:
			"tru_flux":tru_flux,
			"tru_rad":tru_rad,
			"tru_g1":tru_g1,
			"tru_g2":tru_g2,
			"tru_g":tru_g,
			"tru_theta":tru_theta,
			"tru_sersicn":tru_sersicn,
			"tru_mag":tru_mag,
		}
		
		out.update(self.draw_s()) # Here the shear gets drawn for each "galaxy"
		out.update(self.draw_psf()) # Idem, random for each galaxy
		out.update(self.draw_constants())
		
		return out
		

	
	
class EuclidLike_statshear(EuclidLike):
	"""
	To train weights for shear: constant shear per catalog
	"""
	
	def __init__(self, **kwargs):
		EuclidLike.__init__(self, **kwargs)
		
	def stat(self):
		"""
		Supercedes what draw returns, called once for each catalog
		"""
		out = {}
		out.update(self.draw_s()) # Here, it's called for each catalog, and supercedes the galaxy values.
		out.update(self.draw_psf()) # Idem, one random PSF for each catalog
		out.update(self.draw_constants())
		
		return out
			

sp = EuclidLike_statshear(
	name = "figstamps",
	snc_type = 0,
	shear = 0.0
	)

stampsize = 32

psfcat = momentsml.tools.io.readpickle(os.path.join(config.workdir, "psfcat.pkl"))

gsparams = galsim.GSParams(maximum_fft_size=20320)

# Simulating images
momentsml.sim.run.multi(
	simdir=config.simdir,
	simparams=sp,
	drawcatkwargs={"n":24, "nc":4, "stampsize":stampsize},
	drawimgkwargs={"gsparams":gsparams, "sersiccut":5.0}, 
	psfcat=psfcat, psfselect="random",
	ncat=1, nrea=1, ncpu=5,
	savepsfimg=False, savetrugalimg=False
)

momentsml.meas.run.onsims(
	simdir=config.simdir,
	simparams=sp,
	measdir=config.simmeasdir,
	measfct=measfcts.default,
	measfctkwargs={"stampsize":stampsize, "gain":gain},
	ncpu=5,
	skipdone=False
	)
	

cat = momentsml.meas.avg.onsims(
	measdir=config.simmeasdir, 
	simparams=sp,
	task="group",
	groupcols=measfcts.default_groupcols, 
	removecols=measfcts.default_removecols
)

momentsml.tools.table.keepunique(cat)
momentsml.tools.io.writepickle(cat, os.path.join(config.simmeasdir, sp.name, "groupmeascat.pkl"))

print cat["adamom_x", "adamom_y", "tru_mag", "snr"]





import numpy as np
import tenbilac

import logging
logging.basicConfig(level=logging.DEBUG)


data = tenbilac.utils.readpickle("data.pkl")


net = tenbilac.net.Net(1, [5])
net.setidentity()
net.addnoise(wscale=0.1, bscale=0.1)

traindata = tenbilac.data.Traindata(data["normed_train_obs"], data["normed_train_params"], valfrac=0.5, shuffle=True)

train = tenbilac.train.Training(net, traindata, errfctname="msb")

train.opt(algo="bfgs", mbfrac=1, mbloops=1, maxiter=200)

tenbilac.plot.sumevo(train, filepath="train_msb.png")

train.save("train_msb.pkl", keepdata=True)





"""
Training z = x * y with a single product unit, testing this mult neuron.

Single realziation, no noise in the data, not inverse regression.

"""

import numpy as np
import tenbilac

import logging
logging.basicConfig(level=logging.INFO)


nrea = 1
ncas = 1000

# We prepare the data:

# Range of x and y
xs = np.random.uniform(-5, 5, ncas)
ys = np.random.uniform(-10, 30, ncas)

zs = xs * ys 

inputs = np.array([xs, ys])
inputs = np.tile(inputs, (nrea, 1, 1))

# This is 3D (rea, features=2, case)
#print inputs
#print inputs.shape


targets = np.array([zs])

# This is 2D (feature=1, case)
#print targets
#print targets.shape


inputnormer = tenbilac.data.Normer(inputs, type="sa1")
inputs = inputnormer(inputs)
targetnormer = tenbilac.data.Normer(targets, type="sa1")
targets = targetnormer(targets)

print inputnormer
print targetnormer


dat = tenbilac.data.Traindata(inputs=inputs, targets=targets)

net = tenbilac.net.Net(ni=2, nhs=[-1], no=1, actfctname="iden", oactfctname="iden", multactfctname="iden", inames=["x", "y"], onames=["z"])
#net = tenbilac.net.Net(ni=2, nhs=[5, 5], inames=["x", "y"], onames=["z"])

net.setidentity()

# The exact solution:
#net.layers[0].weights[0,1] = 1.0
#net.layers[1].weights[0] = ((inputnormer.b[0]*inputnormer.b[1])/targetnormer.b[0])
print net.report()

#net.addnoise(multwscale=0.01, wscale=0.01, bscale=0.01)

print net.report()

training = tenbilac.train.Training(net, dat, errfctname="mse", autoplot=False, autoplotdirpath="plots")

training.opt(algo="bfgs", maxiter=50, gtol=1e-8)


print net.report()

outs = targetnormer.denorm(net.predict(inputs))
#outs = net.predict(inputs)

assert nrea==1

zerrs = (outs[0] - np.array([zs]))[0]
print np.mean(zerrs)
print np.std(zerrs)

import matplotlib.pyplot as plt

plt.scatter(xs, ys, c=zerrs, s=80, marker="o", edgecolors="face")
plt.xlabel("x")
plt.ylabel("y")
plt.colorbar()
plt.show()





"""
Classic simobscompa plot
"""
import matplotlib
#matplotlib.use("AGG")
matplotlib.use("pdf")

import momentsml
import momentsmlgreat3
import momentsml.plot
from momentsml.tools.feature import Feature
import matplotlib.pyplot as plt
import numpy as np

from momentsml import tools

import config

import os
import logging
logging.basicConfig(format=config.loggerformat, level=logging.INFO)
logger = logging.getLogger(__name__)


import matplotlib.cm
from mpl_toolkits.axes_grid1 import make_axes_locatable
from matplotlib.ticker import MaxNLocator
from matplotlib.ticker import AutoMinorLocator
from matplotlib.lines import Line2D
from matplotlib.ticker import MaxNLocator
from matplotlib.colors import LinearSegmentedColormap, colorConverter
from matplotlib.ticker import ScalarFormatter
import matplotlib.lines as mlines
import matplotlib.ticker as plticker


from matplotlib import rc
rc('font',**{'family':'sans-serif','sans-serif':['Helvetica']})
## for Palatino and other serif fonts use:
#rc('font',**{'family':'serif','serif':['Palatino']})
rc('text', usetex=True)

def _contour(ax, x, y, color="black", bins=30, minline=0.5, maxline=4.0, nlines=3, smooth=2.0, **kwargs):
	"""
	Note that this accepts numpy arrays, not a catalog and features.
	Write a contour wrapper that accepts same input as the others!
	"""
	
	range = [[x.min(), x.max()], [y.min(), y.max()]]
	
	# Choose the default "sigma" contour levels.
	levels = 1.0 - np.exp(-0.5 * np.linspace(minline, maxline, nlines) ** 2)

	H, X, Y = np.histogram2d(x.flatten(), y.flatten(), bins=bins, range=range)

	from scipy.ndimage import gaussian_filter
	if smooth is not None:
		 if gaussian_filter is None:
			 raise ImportError("Please install scipy for smoothing")
		 H = gaussian_filter(H, smooth)

	# Compute the density levels.
	Hflat = H.flatten()
	inds = np.argsort(Hflat)[::-1]
	Hflat = Hflat[inds]
	sm = np.cumsum(Hflat)
	sm /= sm[-1]
	V = np.empty(len(levels))
	for i, v0 in enumerate(levels):
		try:
			V[i] = Hflat[sm <= v0][-1]
		except:
			V[i] = Hflat[0]

	# Compute the bin centers.
	X1, Y1 = 0.5 * (X[1:] + X[:-1]), 0.5 * (Y[1:] + Y[:-1])

	# Extend the array for the sake of the contours at the plot edges.
	H2 = H.min() + np.zeros((H.shape[0] + 4, H.shape[1] + 4))
	H2[2:-2, 2:-2] = H
	H2[2:-2, 1] = H[:, 0]
	H2[2:-2, -2] = H[:, -1]
	H2[1, 2:-2] = H[0]
	H2[-2, 2:-2] = H[-1]
	H2[1, 1] = H[0, 0]
	H2[1, -2] = H[0, -1]
	H2[-2, 1] = H[-1, 0]
	H2[-2, -2] = H[-1, -1]
	X2 = np.concatenate([
		X1[0] + np.array([-2, -1]) * np.diff(X1[:2]),
		X1,
		X1[-1] + np.array([1, 2]) * np.diff(X1[-2:]),
	])
	Y2 = np.concatenate([
		Y1[0] + np.array([-2, -1]) * np.diff(Y1[:2]),
		Y1,
		Y1[-1] + np.array([1, 2]) * np.diff(Y1[-2:]),
	])

	# Plot the contours
	#ax.contour(X2, Y2, H2.T, V, colors=color, **kwargs) # Complains about repeated and non-sorted V values, so here is a quick fix, putting them in the right order:
	V = sorted(list(set(V)))
	ax.contour(X2, Y2, H2.T, V, colors=color, **kwargs)


def simobs(ax, obscat, simcatg3, simcattrain, featx, featy, sidehists=True, sidehistkwargs=None, title=None, legend=False, plotpoints=True, **kwargs):
	"""
	A scatter plot overplotting simulations (in red) and observations (in blue, like the sky).
	Previously the observations were green (like nature), but blue is better for most colorblind people.
			
	:param ax: a matplotlib Axes object
	:param simcat: simulation catalog
	:param obscat: observation catalog
	:param featx: a Feature object telling me what to draw on my x axis
	:param featy: idem for y
	:param sidehists: set this to False if you don't want side histograms
	:param sidehistkwargs: keyword arguments passed to the side hists
	:param title: the title to place on top of the axis.
		The reason why we do not leave this to the user is that the placement changes when sidehists is True.
	:param legend: if True, it writes a self-styled non-invasive "legend" in the top right corner
	
	All further **kwargs** are passed to contour() to make the contours.
		
	"""
	obsdata = tools.feature.get1Ddata(obscat, [featx, featy] , keepmasked=False)	
	simdatag3 = tools.feature.get1Ddata(simcatg3, [featx, featy], keepmasked=False)
	simdatatrain = tools.feature.get1Ddata(simcattrain, [featx, featy], keepmasked=False)


	
	# And the contours:
	contourkwargs = {"bins":100, "smooth":4.0, "minline":1.0, "maxline":2.0, "nlines":2, "zorder":100}
	contourkwargs.update(kwargs)
	_contour(ax, obsdata[featx.colname], obsdata[featy.colname], color="red", linestyles='-', **contourkwargs)
	_contour(ax, simdatag3[featx.colname], simdatag3[featy.colname], color="blue", linestyles=':', **contourkwargs)
	_contour(ax, simdatatrain[featx.colname], simdatatrain[featy.colname], color="green", linestyles='--', **contourkwargs)
	
	
	if sidehists:
	
		# By default, we want to limit the "binning" of the actual histograms (not just their display) to the specified ranges.
		# However, this fails when the "low" or "high" are set to None. Hence some explicit code:
		if featx.low is not None and featx.high is not None: 
			histxrange = (featx.low, featx.high)
		else:
			histxrange = None
		if featy.low is not None and featy.high is not None: 
			histyrange = (featy.low, featy.high)
		else:
			histyrange = None
		# If you do not like this default behaviour, you can overwrite it using the sidehistkwarg "range=None" !
	
		# We now prepare the kwargs for calling hist:
		if sidehistkwargs is None:
			sidehistkwargs = {}
		mysidehistxkwargs = {"histtype":"step", "normed":"True", "bins":50, "alpha":1, "range":histxrange} # for x
		mysidehistxkwargs.update(sidehistkwargs)
		mysidehistykwargs = {"histtype":"step", "normed":"True", "bins":50, "alpha":1, "range":histyrange} # for y
		mysidehistykwargs.update(sidehistkwargs)
	
		divider = make_axes_locatable(ax)
		axhistx = divider.append_axes("top", 1.0, pad=0.1, sharex=ax)
		axhisty = divider.append_axes("right", 1.0, pad=0.1, sharey=ax)
		
		axhistx.hist(obsdata[featx.colname], color="blue", ec="red", ls="-", **mysidehistxkwargs)	
		axhistx.hist(simdatag3[featx.colname], color="red", ec="blue", ls=":", **mysidehistxkwargs)
		axhistx.hist(simdatatrain[featx.colname], color="green", ec="green", ls="--", **mysidehistxkwargs)
		
		axhisty.hist(obsdata[featy.colname], color="blue", ec="red", ls="-", orientation='horizontal', **mysidehistykwargs)
		axhisty.hist(simdatag3[featy.colname], color="red", ec="blue", ls=":", orientation='horizontal', **mysidehistykwargs)
		axhisty.hist(simdatatrain[featy.colname], color="green", ec="green", ls="--", orientation='horizontal', **mysidehistykwargs)
		
		# Hiding the ticklabels
		for tl in axhistx.get_xticklabels():
			tl.set_visible(False)
		for tl in axhisty.get_yticklabels():
			tl.set_visible(False)

		# Hide the hist ticks
		axhistx.yaxis.set_ticks([]) # or set_ticklabels([])
		axhisty.xaxis.set_ticks([])
	
	
	
	# We set the limits and labels:
	ax.set_xlim(featx.low, featx.high)
	ax.set_ylim(featy.low, featy.high)
	ax.set_xlabel(featx.nicename)
	ax.set_ylabel(featy.nicename)
	
	# We want minor ticks:
	ax.xaxis.set_minor_locator(AutoMinorLocator(5))
	ax.yaxis.set_minor_locator(AutoMinorLocator(5))
	
	"""
	if legend:
		ax.annotate("Simulations", color="red", xy=(1.0, 1.0), xycoords='axes fraction', xytext=(-8, -8), textcoords='offset points', ha='right', va='top')
		ax.annotate("Observations", color="blue", xy=(1.0, 1.0), xycoords='axes fraction', xytext=(-8, -24), textcoords='offset points', ha='right', va='top')
	"""
	



def main():

	assert len(config.great3.subfields) == 1
	subfield = config.great3.subfields[0]
	
			

	plotpath = config.great3.path("fig_1_{}_{}.pdf".format(config.great3.get_branchacronym(), subfield))
		
	measdir = config.great3.subpath(subfield, "simmeas")

	simcatg3 = momentsml.tools.io.readpickle(os.path.join(measdir, "simobscompa-G3", "groupmeascat.pkl"))
	#print momentsml.tools.table.info(simcat)
	simcattrain = momentsml.tools.io.readpickle(os.path.join(measdir, "simobscompa-train", "groupmeascat.pkl"))
	#print momentsml.tools.table.info(simcat)

	obscat = momentsml.tools.io.readpickle(config.great3.subpath(subfield, "obs", "img_meascat.pkl"))
	#print momentsml.tools.table.info(obscat)
	#obscat = obscat[:1000]
			
	for cat in [simcatg3, simcattrain, obscat]:
		cat["adamom_log_flux"] = np.log10(cat["adamom_flux"])
		cat["adamom_g"] = np.hypot(cat["adamom_g1"], cat["adamom_g2"])
	

	fig = plt.figure(figsize=(9, 3))
	#fig = plt.figure(figsize=(8, 8))
	
	subfieldstr = str(subfield)
	if len(subfieldstr) == 4: # We cut the "1000"
		subfieldstr = subfieldstr[1:]
	
	fig.text(.02, .58, config.great3.get_branchacronym().upper(), fontdict={"fontsize":22})
	fig.text(.02, .5, "Subfield {}".format(subfieldstr), fontdict={"fontsize":10})

	#fig.text(.05, .7, "GREAT3".format(subfieldstr), fontdict={"fontsize":10, "color":"red"})
	#fig.text(.05, .65, "Mock".format(subfieldstr), fontdict={"fontsize":10, "color":"blue"})
	#fig.text(.05, .6, "Uniform".format(subfieldstr), fontdict={"fontsize":10, "color":"green"})

	
	ax = plt.axes([.3, .2, .29, .78])
	#ax = fig.add_subplot(1, 2, 1)
	if config.great3.obstype == "space":
		simobs(ax, obscat, simcatg3, simcattrain, Feature("adamom_log_flux", 0.3, 2.5, nicename=r"$log_{10}(\mathtt{adamom\_flux})$"), Feature("adamom_sigma", 0, 10, nicename=r"$\mathtt{adamom\_sigma} [\mathrm{pix}]$"), legend=True, plotpoints=False)
	elif config.great3.obstype == "ground":
		simobs(ax, obscat, simcatg3, simcattrain, Feature("adamom_log_flux", 0.3, 2.5, nicename=r"$log_{10}(\mathtt{adamom\_flux})$"), Feature("adamom_sigma", 1., 3., nicename=r"$\mathtt{adamom\_sigma} [\mathrm{pix}]$"), legend=True, plotpoints=False)


	#ax = fig.add_subplot(1, 2, 2)
	ax = plt.axes([.7, .2, .29, .78])
	simobs(ax, obscat, simcatg3, simcattrain, Feature("adamom_g", -0.09, 0.6, nicename=r"$\mathtt{adamom\_g}$"), Feature("adamom_rho4", 1.8, 2.8, nicename=r"$\mathtt{adamom\_rho4}$"), plotpoints=False)
	ax.xaxis.set_major_locator(plticker.MultipleLocator(base=0.2))


	obsline = mlines.Line2D([], [], color='red', ls="-", marker=None, label='GREAT3')
	simg3line = mlines.Line2D([], [], color='blue', ls=":", marker=None, label='Mock sims')
	simtrainline = mlines.Line2D([], [], color='green', ls="--", marker=None, label='Uniform sims')
	
	#plt.legend(handles=[obsline])
	plt.legend(handles=[obsline, simg3line, simtrainline], bbox_to_anchor=(0.16, 0.999), bbox_transform=plt.gcf().transFigure, fontsize=10)
		   
	#plt.tight_layout()

	
	plt.savefig(plotpath)
	

if __name__ == "__main__":
    main()
	



"""
General utility stuff for GREAT3
"""

import os

class Branch:
	"""
	A class to define the file paths of the GREAT3 data and other branch-specific stuff.
	This is to tell code where GREAT3 stuff is.
	It's NOT to tell code where to write (intermediary) files or to specify workdirs etc!
	So in fact there should be no MomentsML related things in here.
	"""

	def __init__(self, experiment, obstype, sheartype, datadir=".", truthdir="."):
	
		assert experiment in ['control', 'real_galaxy', 'variable_psf', 'multiepoch', 'full']
		assert obstype in ['ground', 'space']
		assert sheartype in ['constant', 'variable']
		
		self.experiment = experiment
		self.obstype = obstype
		self.sheartype = sheartype
	
		self.datadir = datadir
		self.truthdir = truthdir 
		
		
	
	def branchtuple(self):
		"""
		The branch codes in form of a tuple
		"""
		return(self.experiment, self.obstype, self.sheartype)
	
	def branchcode(self):
		"""
		The branch codes in form of a tuple
		"""
		return "%s-%s-%s" % (self.experiment, self.obstype, self.sheartype)
	
	def get_branchacronym(self):
		"""
		Get the acrnonym of the branch (e.g. cgv)
		"""
		return "".join([self.experiment[0], self.obstype[0], self.sheartype[0]])
	

	def __str__(self):
		return "Branch (%s, %s, %s)" % self.branchtuple()
	

	def stampsize(self):
		"""
		The stamp size "s" of this branch, in pixels
		"""
		if self.obstype == "ground":
			return 48
		elif self.obstype == "space":
			return 96
		
	def ntiles(self):
		if self.experiment in ['control', 'real_galaxy', 'multiepoch']:
			return 1
		if self.obstype == "ground":
			return 5
		elif self.obstype == "space":
			return 20
		
	def pixelscale(self):
		"""
		The pixel scale of this branch, in pixels
		"""
		if self.obstype == "ground":
			return 0.2
		elif self.obstype == "space":
			if self.experiment=="multiepoch":
				return 0.1
			else:
				return 0.05

	def gain(self):
		"""
		It seems that GREAT3 images are given in scaled electrons, but it doesn't matter as noise is just stationarry Gaussian anyway.
		For SExtractor one should set gain = 0 which corresponds to infinity to get this Gaussian noise right.
		"""
		return 1.0

	
	def branchdir(self, truth=False):
		"""
		Where all the data is
		If truth is True, returns the path within the truthdir instead.
		"""
		if truth is False:
			return os.path.join(self.datadir, "/".join(self.branchtuple()))
		else:
			return os.path.join(self.truthdir, "/".join(self.branchtuple()))


	# Now we only define here the paths to the "input" stuff, set by GREAT3.
	# No MomentsML-internal files!
	
	
	def get_ftiles(self, xt, yt):
		"""
		A little helper for the name of the files generator functions below. It handles the tile id
		"""
		if not (xt is None or yt is None):
			return "-%02d-%02d" % (xt,yt)
		else:
			return ""


	def galimgfilepath(self, subfield, xt=None, yt=None,epoch=0, folder=None):
		if folder==None:
			folder=self.branchdir()
		return os.path.join(folder, "image-%03i-%i%s.fits" % (subfield, epoch, self.get_ftiles(xt,yt))) # This is set by GREAT3


	def starimgfilepath(self, subfield, xt=None, yt=None, epoch=0):
		return os.path.join(self.branchdir(), "starfield_image-%03i-%i%s.fits" % 
						(subfield, epoch, self.get_ftiles(xt,yt))) # This is set by GREAT3


	def galcatfilepath(self, subfield, xt=None, yt=None, folder=None):
		if folder==None:
			folder=self.branchdir()

		fname="galaxy_catalog-%03i%s.txt" % (subfield,self.get_ftiles(xt,yt))
		return os.path.join(folder, fname) # This is set by GREAT3

		
	def starcatfilepath(self, subfield, xt=None, yt=None, folder=None):
		if folder==None:
			folder=self.branchdir()
		return os.path.join(folder, 'star_catalog-%03i%s.txt' % \
						(subfield,self.get_ftiles(xt,yt))) # This is set by GREAT3
	
	
	def trushearfilepath(self, subfield, folder=None):
		if folder is None:
			folder = self.branchdir(truth=True)
		return os.path.join(folder, 'shear_params-%03i.txt' % (subfield))
	
	def trustarfilepath(self, subfield, folder=None):
		"""
		The psf g1 g2, used to "rotate" metrics
		"""
		if folder is None:
			folder = self.branchdir(truth=True)
		return os.path.join(folder, 'starshape_parameters-%03i-0.txt' % (subfield))
	


import momentsml.tools.calc
import numpy as np

def metrics(cat, trucols, predcols, psfgcols, latex=True):
	"""
	Computes GREAT3 metrics m and c for 1, 2, + and x, including errors.
	trucols and predcols are tuples with column names, e.g. ("tru_s1", "tru_s2")
	
	It has been checked that results are identical to what the public great3 code gives (on the "simple" branches only).
	"""
	
	returndict = {}
	
	assert cat[trucols[0]].ndim == 1
	assert cat[predcols[1]].ndim == 1

	
	pixres1 = momentsml.tools.calc.linreg(cat[trucols[0]], cat[predcols[0]], prob=0.68) # Component 1
	pixres2 = momentsml.tools.calc.linreg(cat[trucols[1]], cat[predcols[1]], prob=0.68) # Component 2
	
	returndict["m1"] = pixres1["m"]
	returndict["m1err"] = pixres1["merr"]
	returndict["c1"] = pixres1["c"]
	returndict["c1err"] = pixres1["cerr"]
	
	returndict["m2"] = pixres2["m"]
	returndict["m2err"] = pixres2["merr"]
	returndict["c2"] = pixres2["c"]
	returndict["c2err"] = pixres2["cerr"]
	
	
	# Now the rotated metrics
	# This simply follows great3-public/metrics/evaluate
		
	# angles:
	rotations = .5 * np.arctan2(cat[psfgcols[1]], cat[psfgcols[0]]) # Note the inverted order of components
		
	# rotating predictions:
	g1prot = cat[predcols[0]] * np.cos(-2. * rotations) - cat[predcols[1]] * np.sin(-2. * rotations)
	g2prot = cat[predcols[0]] * np.sin(-2. * rotations) + cat[predcols[1]] * np.cos(-2. * rotations)
	
	# rotating truth:
	g1trot = cat[trucols[0]] * np.cos(-2. * rotations) - cat[trucols[1]] * np.sin(-2. * rotations)
	g2trot = cat[trucols[0]] * np.sin(-2. * rotations) + cat[trucols[1]] * np.cos(-2. * rotations)
 	
	
	pixresplus = momentsml.tools.calc.linreg(g1trot, g1prot, prob=0.68)
	pixrescros = momentsml.tools.calc.linreg(g2trot, g2prot, prob=0.68)
	
	returndict["m+"] = pixresplus["m"]
	returndict["m+err"] = pixresplus["merr"]
	returndict["c+"] = pixresplus["c"]
	returndict["c+err"] = pixresplus["cerr"]

	returndict["mx"] = pixrescros["m"]
	returndict["mxerr"] = pixrescros["merr"]
	returndict["cx"] = pixrescros["c"]
	returndict["cxerr"] = pixrescros["cerr"]



	if latex is True:
		
		kilo = {}
		for (key, value) in returndict.iteritems():
			kilo[key] = 1000.0 * value
		
		ret = ["Branch & ",
			"${r[m1]:+.2f}\pm{r[m1err]:.2f}$ & ".format(r=kilo),
			"${r[m2]:+.2f}\pm{r[m2err]:.2f}$ & ".format(r=kilo),
			"${r[m+]:+.2f}\pm{r[m+err]:.2f}$ & ".format(r=kilo),
			"${r[mx]:+.2f}\pm{r[mxerr]:.2f}$ & ".format(r=kilo),
			"${r[c1]:+.2f}\pm{r[c1err]:.2f}$ & ".format(r=kilo),
			"${r[c2]:+.2f}\pm{r[c2err]:.2f}$ & ".format(r=kilo),
			"${r[c+]:+.2f}\pm{r[c+err]:.2f}$ & ".format(r=kilo),
			"${r[cx]:+.2f}\pm{r[cxerr]:.2f}$ \\".format(r=kilo),
			]
			
		print "LaTeX table line:"
		print "".join(ret)
	
	return returndict
	
	
	
	
	
	




import os

import logging
logger = logging.getLogger(__name__)


def mkdir(somedir):
	"""
	A wrapper around os.makedirs, with some logging.
	:param somedir: a name or path to a directory which I should make.
	"""
	
	raise RuntimeError("I suggest to stop using this, it doesn't even log so I don't see the point.")
	
	if not os.path.isdir(somedir):
		os.makedirs(somedir)



import matplotlib
matplotlib.use("AGG")

import argparse

import momentsml.tools
import momentsml.learn
import momentsml
import tenbilac

import config
import numpy as np
import os
import glob
import shutil



import logging
logging.basicConfig(format=config.loggerformat, level=logging.DEBUG)
logger = logging.getLogger(__name__)



wtraindir = os.path.join(config.traindir, config.datasets["tw"] + "_with_" + config.datasets["tp"] + "_" + config.sconfname)
catpath = os.path.join(wtraindir, "groupmeascat_predforw.pkl")

cat = momentsml.tools.io.readpickle(catpath)


	
# Running the training	
dirnames = momentsml.learn.tenbilacrun.train(cat, config.weightconflist, wtraindir)

print dirnames

#dirnames = ["ada4s1w_sum33wmassshort", "ada4s2w_sum33wmassshort"]




import matplotlib
matplotlib.use("AGG")

import momentsml.tools
import momentsml.learn
import momentsml

import config
import numpy as np
import os


import logging
logging.basicConfig(format=config.loggerformat, level=logging.DEBUG)
logger = logging.getLogger(__name__)

	
# Getting the path to the correct directories
traindir = os.path.join(config.traindir, config.datasets["tp"])
	
# And to the validation catalogue
valcatpath = os.path.join(config.simmeasdir, config.datasets["vp"], "groupmeascat.pkl")
predcatpath = os.path.join(config.valdir, config.valname + ".pkl")


# The prediction:

cat = momentsml.tools.io.readpickle(valcatpath)


predcat = momentsml.learn.tenbilacrun.predict(cat, config.shearconflist, traindir)

momentsml.tools.io.writepickle(predcat, predcatpath)






import numpy as np
import tenbilac

import logging
logging.basicConfig(level=logging.INFO)


ni = 5
ngal = 4
nrea = 3
no = 1
noise_scale = 1.0

# 3D case:
inputs = np.random.randn(ni * ngal * nrea).reshape((nrea, ni, ngal))*1.0 + 5.0

# Modifying one of the features:
inputs[:,2,:] -= 5.0

# 2D case:
#inputs = np.random.randn(ni * ngal).reshape((ni, ngal))*1000.0 + 15.0


print inputs
print inputs.shape


normer = tenbilac.data.Normer(inputs, type="sa1")

print normer

normed = normer(inputs)

print normed

print np.min(normed), np.max(normed)

denormed = normer.denorm(normed)

if not np.allclose(denormed, inputs):
	raise RuntimeError("Comparision test failed !")
else:
	print "Test is good!"




#import matplotlib
#matplotlib.use("AGG")

import momentsml
import momentsml.tools
import momentsml.plot
from momentsml.tools.feature import Feature

import momentsmlgreat3
import astropy


import config
import numpy as np
import os
import matplotlib.pyplot as plt


import logging
logging.basicConfig(format=config.loggerformat, level=logging.DEBUG)
logger = logging.getLogger(__name__)


catpath = config.great3.path("{}_summary_{}.pkl".format(config.datasets["mimic-great3"], config.predcode))
cat = momentsml.tools.io.readpickle(catpath)


mets = momentsmlgreat3.utils.metrics(cat, ("tru_s1", "tru_s2"), ("pre_s1w", "pre_s2w"), psfgcols=("tru_psf_g1", "tru_psf_g2"))

